import logging
import numpy as np
import scipy.interpolate as spi

import allensdk.model.glif.glif_neuron as glif_neuron

# TODO: license
# TODO: document

class GlifNeuronException( Exception ):
    """ Exception for catching simulation errors and reporting intermediate data. """
    def __init__(self, message, data):
        super(Exception, self).__init__(message)
        self.data = data

class GlifBadInitializationException( Exception ):
    """ Exception raised when voltage is above threshold at the beginning of a sweep. i.e. probably caused by the optimizer. """
    def __init__(self, message, dv, step):
        super(Exception, self).__init__(message)
        self.dv = dv
        self.step=step

class GlifOptimizerNeuron( glif_neuron.GlifNeuron ):
    '''Contains methods for running the neuron model in a "forced-spike" paradigm
    used during optimization.
    '''
    
    TYPE = "GLIF"
    
    def __init__(self, *args, **kwargs):
        
        super(GlifOptimizerNeuron, self).__init__(*args, **kwargs)
        
        self.extrapolation_method_name = kwargs.get('extrapolation_method_name', 'endpoints')
        if self.extrapolation_method_name == 'endpoints':
            self.extrapolation_method = extrapolate_model_spike_from_endpoints
        elif self.extrapolation_method_name == 'endpoints_single_tau':
            self.extrapolation_method = extrapolate_model_spike_from_endpoints_single_tau
        else:
            raise Exception('unknown extrapolation method: %s' % self.extrapolation_method_name)
        
        #TODO: what is this where is it comming from?
        self.dt_multiplier = kwargs.get('dt_multiplier', 1)
        self.El_reference = kwargs.get('El_reference',None)
            
            
    @classmethod
    def from_dict(cls, d):
        
        return cls(El = d['El'],
                   dt = d['dt'],
#                   tau = d['tau'],
                   asc_tau_array=d['asc_tau_array'],
                   R_input = d['R_input'],
                   C = d['C'],
                   asc_amp_array = d['asc_amp_array'],
                   spike_cut_length = d['spike_cut_length'],
                   th_inf = d['th_inf'],
                   th_adapt=None,
                   coeffs = d.get('coeffs', {}),
                   AScurrent_dynamics_method = d['AScurrent_dynamics_method'],
                   voltage_dynamics_method = d['voltage_dynamics_method'],
                   threshold_dynamics_method = d['threshold_dynamics_method'],
                   voltage_reset_method = d['voltage_reset_method'],
                   AScurrent_reset_method = d['AScurrent_reset_method'],
                   threshold_reset_method = d['threshold_reset_method'],
                   init_voltage = d['init_voltage'],
                   init_threshold = d['init_threshold'],
                   init_AScurrents = d['init_AScurrents'],
                   extrapolation_method_name = d.get('extrapolation_method_name', 'endpoints'),
                   dt_multiplier = d.get('dt_multiplier',1),
                   El_reference = d['El_reference']
                   )
        
    @classmethod
    def from_dict_legacy(cls, d):
        
        return cls(El = d['El'],
                   dt = d['dt'],
#                   tau = d['tau'],
                   asc_tau_array=d['asc_tau_array'],
                   R_input = d['R_input'],
                   C = d['C'],
                   asc_amp_array = d['asc_amp_array'],
                   spike_cut_length = d['spike_cut_length'],
                   th_inf = d['th_inf'],
                   th_adapt=d['th_adapt'],
                   coeffs = d.get('coeffs', {}),
                   AScurrent_dynamics_method = d['AScurrent_dynamics_method'],
                   voltage_dynamics_method = d['voltage_dynamics_method'],
                   threshold_dynamics_method = d['threshold_dynamics_method'],
                   voltage_reset_method = d['voltage_reset_method'],
                   AScurrent_reset_method = d['AScurrent_reset_method'],
                   threshold_reset_method = d['threshold_reset_method'],
                   init_voltage = d['init_voltage'],
                   init_threshold = d['init_threshold'],
                   init_AScurrents = d['init_AScurrents'],
                   extrapolation_method_name = d.get('extrapolation_method_name', 'endpoints'),
                   dt_multiplier = d.get('dt_multiplier',1)
                   )
        
    def to_dict(self):
        
        curr_dict = super(GlifOptimizerNeuron, self).to_dict()
        curr_dict.update({'extrapolation_method_name':self.extrapolation_method_name,
                          'dt_multiplier':self.dt_multiplier,
                          'El_reference':self.El_reference})

        return curr_dict

    def run_with_biological_spikes(self, stimulus, response, bio_spike_time_steps):
        """ Run the neuron simulation over a stimulus, but do not allow the model to spike on its own.  Rather,
        force the simulation to spike and reset at a given set of spike indices.  Dynamics rules are applied
        between spikes regardless of the simulated voltage and threshold values.  Reset rules are applied only 
        at input spike times. This is used during optimization to force the model to follow the spikes of biological data.
        The model is optimized in this way so that history effects due to spiking can be adequately modeled.  For example, 
        every time the model spikes a new set of afterspike currents will be initiated. To ensure that afterspike currents 
        can be optimized, we force them to be initiated at the time of the biological spike.

        Parameters
        ----------
        stimulus : np.ndarray
            vector of scalar current values
        respones : np.ndarray
            vector of scalar voltage values
        bio_spike_time_steps : list
            spike time step indices

        Returns
        -------
        dict
            a dictionary containing:
                'voltage': simulated voltage values,
                'threshold': simulated threshold values,
                'AScurrent_matrix': afterspike currents during the simulation,
                'grid_model_spike_times': spike times of the model aligned to the simulation grid (when it would have spiked),
                'interpolated_model_spike_times': spike times of the model linearly interpolated between time steps,
                'grid_ISI': interspike interval between grid model spike times,
                'interpolated_ISI': interspike interval between interpolated model spike times,
                'grid_bio_spike_model_voltage': voltage of the model at biological/input spike times,
                'grid_bio_spike_model_threshold': voltage of the model at biological/input spike times interpolated between time steps
        """
        
        self.threshold_components = None  #get rid of lingering threshold components

        voltage_t0 = self.init_voltage
        threshold_t0 = self.init_threshold
        AScurrents_t0 = self.init_AScurrents
        
        if voltage_t0>threshold_t0:
            raise GlifBadInitializationException("Voltage STARTS above threshold: voltage_t0 (%f) threshold_t0 (%f)" % ( voltage_t0, threshold_t0, voltage_t0 - threshold_t0, 10000000.0))

        start_index = 0
        end_index = 0
            
        try:
            num_spikes = len(bio_spike_time_steps)
            
            # if there are no target spikes, just run until the model spikes
            if num_spikes == 0:  

                start_index = 0
                end_index = len(stimulus)

                # evaluate the model starting from the beginning until the model spikes
                run_data = self.run_until_biological_spike(voltage_t0, threshold_t0, AScurrents_t0, 
                                                           stimulus, response, start_index, end_index, 
                                                           []) 

                voltage = run_data['voltage']
                threshold = run_data['threshold']
                AScurrent_matrix = run_data['AScurrent_matrix']

                if len(voltage) != len(stimulus):
                    logging.warning('Your voltage output is not the same length as your stimulus')
                if len(threshold) != len(stimulus):
                    logging.warning('Your threshold output is not the same length as your stimulus')
                if len(AScurrent_matrix) != len(stimulus):
                    logging.warning('Your AScurrent_matrix output is not the same length as your stimulus')
                                          
                # do not keep track of the spikes in the model that spike if the target doesn't spike.
                grid_ISI = np.array([])
                interpolated_ISI = np.array([])

                grid_model_spike_times = np.array([]) 
                interpolated_model_spike_times = np.array([])
                
                grid_model_spike_voltages = np.array([]) 
                interpolated_model_spike_voltages = np.array([]) 

                grid_bio_spike_model_voltage = np.array([])
                grid_bio_spike_model_threshold = np.array([])
            else:
                # initialize the output arrays
                grid_ISI = np.empty(num_spikes)
                interpolated_ISI = np.empty(num_spikes)

                grid_model_spike_times = np.empty(num_spikes)
                interpolated_model_spike_times = np.empty(num_spikes)
                
                grid_model_spike_voltages = np.empty(num_spikes) 
                interpolated_model_spike_voltages = np.empty(num_spikes)

                grid_bio_spike_model_voltage = np.empty(num_spikes)
                grid_bio_spike_model_threshold = np.empty(num_spikes)

                spikeIndStart = 0
                
                voltage = np.empty(len(stimulus))
                voltage[:] = np.nan  
                threshold = np.empty(len(stimulus))
                threshold[:] = np.nan
                AScurrent_matrix = np.empty(shape=(len(stimulus), len(AScurrents_t0)))
                AScurrent_matrix[:] = np.nan
        
                # run the simulation over the interspike intervals (starting at the beginning of the simulation). 
                start_index = 0
                for spike_num in range(num_spikes):

                    if spike_num % 10 == 0:
                        logging.debug("spike %d / %d" % (spike_num,  num_spikes))

                    end_index = int(bio_spike_time_steps[spike_num])
    
                    assert start_index < end_index, Exception("start_index > end_index: this is probably because spike_cut_length is longer than the previous inter-spike interval")

                    # run the simulation over this interspike interval
#                     t0 = time.time()
                    run_data = self.run_until_biological_spike(voltage_t0, threshold_t0, AScurrents_t0, 
                                                               stimulus, response, start_index, end_index, 
                                                               bio_spike_time_steps)
#                     print 'fast', time.time() - t0
                    
#                     curr_voltage = run_data_fast['voltage']
#                     curr_threshold = run_data_fast['threshold']
#                     voltage_scrubbed = curr_voltage[np.logical_not(np.isnan(curr_voltage))]
#                     threshold_scrubbed = curr_threshold[np.logical_not(np.isnan(curr_threshold))]  
#                     
#                     tmp_t = np.linspace(0,1,len(voltage_scrubbed))
#                     print voltage_scrubbed
#                     plt.plot(tmp_t, voltage_scrubbed)
#                     plt.plot(tmp_t, threshold_scrubbed)
#                     
#                     plt.show()
#                     sys.exit()

#                     for key, val in run_data.items():
#                         print key, val
#                     sys.exit()

                    
                    # assign the simulated data to the correct locations in the output arrays
                    voltage[start_index:end_index] = run_data['voltage']
                    threshold[start_index:end_index] = run_data['threshold']
                    AScurrent_matrix[start_index:end_index,:] = run_data['AScurrent_matrix']

                    grid_ISI[spike_num] = run_data['grid_model_spike_time']
                    interpolated_ISI[spike_num] = run_data['interpolated_model_spike_time']

                    grid_model_spike_times[spike_num] = run_data['grid_model_spike_time'] + start_index * self.dt 
                    interpolated_model_spike_times[spike_num] = run_data['interpolated_model_spike_time'] + start_index * self.dt
                    
                    grid_model_spike_voltages[spike_num] = run_data['grid_model_spike_voltage']
                    interpolated_model_spike_voltages[spike_num] = run_data['interpolated_model_spike_voltage'] 

                    grid_bio_spike_model_voltage[spike_num] = run_data['grid_bio_spike_model_voltage']
                    grid_bio_spike_model_threshold[spike_num] = run_data['grid_bio_spike_model_threshold']

                    # update the voltage, threshold, and afterspike currents for the next interval
                    voltage_t0 = run_data['voltage_t0']
                    threshold_t0 = run_data['threshold_t0']
                    AScurrents_t0 = run_data['AScurrents_t0']
                    
                    start_index = end_index

                    # if cutting spikes, jump forward the appropriate amount of time
                    if self.spike_cut_length > 0:
                        start_index += self.spike_cut_length
                
                # simulate the portion of the stimulus between the last spike and the end of the array.
                # no spikes are recorded from this time!
                run_data = self.run_until_biological_spike(voltage_t0, threshold_t0, AScurrents_t0, 
                                                           stimulus, response, start_index, len(stimulus),
                                                           bio_spike_time_steps) 

                voltage[start_index:] = run_data['voltage']
                threshold[start_index:] = run_data['threshold']
                AScurrent_matrix[start_index:,:] = run_data['AScurrent_matrix']

            # make sure that the output data has the correct number of spikes in it
            if ( len(interpolated_model_spike_times) != num_spikes or 
                 len(grid_model_spike_times) != num_spikes or 
                 len(grid_ISI) != num_spikes or 
                 len(interpolated_ISI) != num_spikes or 
                 len(grid_bio_spike_model_voltage) != num_spikes or 
                 len(grid_bio_spike_model_threshold) != num_spikes):
                raise Exception('The number of spikes in your output does not match your target')

        except GlifNeuronException, e:
            
            # if an exception was raised during run_until_spike, record any simulated data before exiting
            voltage[start_index:end_index] = e.data['voltage']
            threshold[start_index:end_index] = e.data['threshold']
            AScurrent_matrix[start_index:end_index,:] = e.data['AScurrent_matrix']

            out = {
                'voltage': voltage,
                'threshold': threshold,
                'AScurrent_matrix': AScurrent_matrix,

                'grid_ISI': grid_ISI,
                'interpolated_ISI': interpolated_ISI,

                'grid_model_spike_times': grid_model_spike_times,
                'interpolated_model_spike_times': interpolated_model_spike_times,
                
                'grid_model_spike_voltages': grid_model_spike_voltages,
                'interpolated_model_spike_voltages': interpolated_model_spike_voltages,

                'grid_bio_spike_model_voltage': grid_bio_spike_model_voltage,
                'grid_bio_spike_model_threshold': grid_bio_spike_model_threshold 
                }

            raise GlifNeuronException(e.message, out)

        return {
            'voltage': voltage,
            'threshold': threshold,
            'AScurrent_matrix': AScurrent_matrix,

            'grid_model_spike_times': grid_model_spike_times,
            'interpolated_model_spike_times': interpolated_model_spike_times,
            
            'grid_model_spike_voltages': grid_model_spike_voltages,
            'interpolated_model_spike_voltages': interpolated_model_spike_voltages,

            'grid_ISI': grid_ISI,
            'interpolated_ISI': interpolated_ISI,

            'grid_bio_spike_model_voltage': grid_bio_spike_model_voltage,
            'grid_bio_spike_model_threshold': grid_bio_spike_model_threshold
        }
        
    def run_until_biological_spike(self, voltage_t0, threshold_t0, AScurrents_t0, 
                                   stimulus, response, start_index, after_end_index, 
                                   bio_spike_time_steps):
        """ Run the neuron simulation over a segment of a stimulus given initial conditions for use in the "forced spike"
        optimization paradigm. [Note: the section of stimulus 
        is meant to be between two biological neuron spikes.  Thus the stimulus is during the interspike interval (ISI)].  The 
        model is simulated until either the model spikes or the end of the segment is reached.  If the model does not spike, a 
        spike time is extrapolated past the end of the simulation segment.  
        
        This function also returns the initial conditions for the subsequent stimulus segment. In the forced spike paradigm 
        there are several ways 

        Parameters
        ----------
        voltage_t0 : float
            the current voltage of the neuron
        threshold_t0 : float
            the current spike threshold level of the neuron
        AScurrents_t0 : np.ndarray
            the current state of the afterspike currents in the neuron
        stimulus : np.ndarray
            the full stimulus array (not just the segment of data being simulated)
        response : np.ndarray
            the full response array (not just the segment of data being simulated)
        start_index : int
            index of global stimulus at which to start simulation
        after_end_index : int
            index of global stimulus *after* the last index to be simulated
        bio_spike_time_steps : list
            time steps of input spikes

        Returns
        -------
        dict
            a dictionary containing:
                'voltage': simulated voltage value
                'threshold': simulated threshold values
                'AScurrent_matrix': afterspike current values during the simulation
                'grid_model_spike_time': model spike time (in units of dt) 
                'interpolated_model_spike_time': model spike time (in units of dt) interpolated between time steps
                'voltage_t0': reset voltage value to be used in subsequent simulation interval 
                'threshold_t0': reset threshold value to be used in subsequent simulation interval
                'AScurrents_t0': reset afterspike current value to be used in subsequent simulation interval
                'grid_bio_spike_model_voltage': model voltage at the time of the input spike
                'grid_bio_spike_model_threshold': model threshold at the time of the input spike
        """
        
        grid_model_spike_time = None
        grid_model_spike_voltage = None
        interpolated_model_spike_time = None
        interpolated_model_spike_voltage = None

        # preallocate arrays and matricies
        num_time_steps_fine = after_end_index - start_index
        t_fine_grid = np.arange(num_time_steps_fine)*self.dt

        #--------------------------------------------------------------------------------
        #---Apply refinement factor to integrate over larger time steps (assumes--------- 
        #---current within the steps can be averaged):----------------------------------
        #--------------------------------------------------------------------------------
        dt_old = self.dt
        self.dt =  self.dt*self.dt_multiplier
        
        # define the local course grain indicies note the last graining will be shorter and is appended to the end
        local_coarse_indicies=np.append(np.arange(num_time_steps_fine)[::self.dt_multiplier], after_end_index - start_index) #the last indicie in this array is still one longer than the last simulated index 
        # convert the local coarse grained indicies into global incidies 
        global_coarse_indicies=local_coarse_indicies+start_index

        # TODO: I dont think this does anything.
        if len(local_coarse_indicies)==2:  
            pass
        
        num_time_steps_coarse = len(local_coarse_indicies)
        voltage_out_coarse_grid = np.empty(num_time_steps_coarse)
        voltage_out_coarse_grid[:] = np.nan
        threshold_out_coarse_grid = np.empty(num_time_steps_coarse)
        threshold_out_coarse_grid[:] = np.nan
        AScurrent_matrix_coarse_grid = np.empty(shape=(num_time_steps_coarse, len(AScurrents_t0))) 
        AScurrent_matrix_coarse_grid[:] = np.nan
        # these grid times are in the local frame of reference
        t_coarse_grid = np.arange(num_time_steps_coarse-1)*self.dt #subtracting the one off here because appending the actual last time that is not the same dt.
        t_coarse_grid = np.append(t_coarse_grid, t_fine_grid[-1])
        dt_vector=t_coarse_grid[1:]-t_coarse_grid[:-1] #note that this vector is one index shorter than the t_course_grid

        # Define the coarse grain stimulus by taking the stimulus average between indicies.  Note that since the initial input voltage is recorded in 
        # the output vectors the stimulus average is indeed the input to the correct time step (i.e. current being fed in is the average current before the step)
        stimulus_coarse=[stimulus[global_coarse_indicies[ii]:global_coarse_indicies[ii+1]].mean() for ii in range(len(global_coarse_indicies)-1)]

        # step though time steps and calculate voltage values
        for time_step in range(len(local_coarse_indicies)-1):  #minus 1 is needed to match vector sizes because initial inputs are recorded in output vectors.
            # update output values (Note: in general one can update values before or after the first time step.  
            # Here the input starting value of voltage is recorded before a time step. This means the last value is not recorded.)
            voltage_out_coarse_grid[time_step] = voltage_t0 
            threshold_out_coarse_grid[time_step] = threshold_t0
            AScurrent_matrix_coarse_grid[time_step,:] = np.matrix(AScurrents_t0) 
            
            # record error in optimization if they are happening
            if np.isnan(voltage_t0) or np.isinf(voltage_t0) or np.isnan(threshold_t0) or np.isinf(threshold_t0) or any(np.isnan(AScurrents_t0)) or any(np.isinf(AScurrents_t0)):
                logging.error(self)
                logging.error('time step: %d / %d' % (time_step, num_time_steps_coarse))
                logging.error('    voltage_t0: %f' % voltage_t0)
                logging.error('    voltage started the run at: %f' % voltage_out_coarse_grid[0])
                logging.error('    voltage before: %s' % voltage_out_coarse_grid[time_step-20:time_step])
                logging.error('    threshold_t0: %f' % threshold_t0)
                logging.error('    threshold started the run at: %f' % threshold_out_coarse_grid[0])
                logging.error('    threshold before: %s' % threshold_out_coarse_grid[time_step-20:time_step])
                logging.error('    AScurrents_t0: %s' % AScurrents_t0)
                if 'a_spike' in self.threshold_dynamics_method.params:
                    logging.error('    a_spike: %s' % self.threshold_dynamics_method.params['a_spike'])
                if 'b_spike' in self.threshold_dynamics_method.params:
                    logging.error('    b_spike: %s' % self.threshold_dynamics_method.params['b_spike'])
                
                # plot output in original index space
                temp_fine_grid_for_intp=np.arange(0,t_coarse_grid[time_step-1], dt_old)
                voltage_out_fine_grid = np.empty(num_time_steps_fine)
                voltage_out_fine_grid[:] = np.nan
                threshold_out_fine_grid = np.empty(num_time_steps_fine)
                threshold_out_fine_grid[:] = np.nan

                fv = spi.interp1d(t_coarse_grid[:time_step], voltage_out_coarse_grid[:time_step], assume_sorted=True, bounds_error=False, fill_value=voltage_out_coarse_grid[-1])
                ft = spi.interp1d(t_coarse_grid[:time_step], threshold_out_coarse_grid[:time_step], assume_sorted=True, bounds_error=False, fill_value=threshold_out_coarse_grid[-1])

                voltage_with_error = fv(temp_fine_grid_for_intp)
                threshold_with_error = ft(temp_fine_grid_for_intp)
                voltage_out_fine_grid[:len(voltage_with_error)]=voltage_with_error
                threshold_out_fine_grid[:len(threshold_with_error)]=threshold_with_error
                
                AScurrent_matrix = np.empty(shape=(num_time_steps_fine, len(AScurrents_t0)))
                AScurrent_matrix[:] = np.nan
                for ii in range(len(AScurrents_t0)):
                    curr_fASc = spi.interp1d(t_coarse_grid[:time_step], AScurrent_matrix_coarse_grid[:time_step,ii], assume_sorted=True, bounds_error=False, fill_value=AScurrent_matrix_coarse_grid[-1,ii])
                    temp_asc=curr_fASc(temp_fine_grid_for_intp)
                    AScurrent_matrix[:len(temp_asc),ii] = temp_asc 
                
                raise GlifNeuronException('Invalid threshold, voltage, or after-spike current encountered.', {                                                                         
                        'voltage': voltage_out_fine_grid,
                        'threshold': threshold_out_fine_grid,
                        'AScurrent_matrix': AScurrent_matrix
                        })    
            
            # changing dt be the dt of the coarse bin (which is variable for the last bin)
            self.dt=dt_vector[time_step]
            (voltage_t1, threshold_t1, AScurrents_t1) = self.dynamics(voltage_t0, threshold_t0, AScurrents_t0, stimulus_coarse[time_step], time_step+start_index, bio_spike_time_steps) #TODO fix list versus array

            # updating the input values 
            voltage_t0=voltage_t1
            threshold_t0=threshold_t1
            AScurrents_t0=AScurrents_t1
            
        # Inserting the last values into the nan at the end of the matricies so that when do the interpolation the end of the vector will not be nans
        # Note this should not mess with any of the outputs because is the interploated values that are the output.
        voltage_out_coarse_grid[time_step+1] = voltage_t0 
        threshold_out_coarse_grid[time_step+1] = threshold_t0
        AScurrent_matrix_coarse_grid[time_step+1,:] = np.matrix(AScurrents_t0)

        # Reset dt to previous value:
        self.dt =  dt_old

        fv = spi.interp1d(t_coarse_grid, voltage_out_coarse_grid, assume_sorted=True, bounds_error=False, fill_value=voltage_out_coarse_grid[-1])
        ft = spi.interp1d(t_coarse_grid, threshold_out_coarse_grid, assume_sorted=True, bounds_error=False, fill_value=threshold_out_coarse_grid[-1])
        voltage_out = fv(t_fine_grid)
        threshold_out = ft(t_fine_grid)
        
        # initalize after spike current matrix
        AScurrent_matrix = np.empty(shape=(num_time_steps_fine, len(AScurrents_t0)))
        AScurrent_matrix[:] = np.nan
        for ii in range(len(AScurrents_t0)):
            curr_fASc = spi.interp1d(t_coarse_grid, AScurrent_matrix_coarse_grid[:,ii], assume_sorted=True, bounds_error=False, fill_value=AScurrent_matrix_coarse_grid[-1,ii])
            AScurrent_matrix[:,ii] = curr_fASc(t_fine_grid) 
        
        # find where model voltage crosses model threshold 
        grid_model_spike_time, grid_model_spike_voltage, interpolated_model_spike_time, interpolated_model_spike_voltage = find_first_model_spike(voltage_out, threshold_out, voltage_t1, threshold_t1, self.dt)
        # if the model never spiked, extrapolate to guess when it would have spiked
        if grid_model_spike_time is None: 
            grid_model_spike_time, grid_model_spike_voltage, interpolated_model_spike_time, interpolated_model_spike_voltage = self.extrapolation_method(self, voltage_out, threshold_out, voltage_t1, threshold_t1, self.dt)
        
        # when the target spikes, reset so that next round will start at reset but not recording it in the voltage here.
        # note that at the last section of the stimulus where there is no current injected the model will be reset even if
        # the biological neuron doesn't spike.  However, this doesnt matter as it won't be recorded. 
        num_spikes = len(bio_spike_time_steps)
        if num_spikes > 0:
            if after_end_index<len(stimulus):
                
                #TODO: ????????????????????????????????WHAT IS THIS?????????????????????????????????????????????????????????????
                #I CANT FIGURE OUT WHY THIS WOULD HAVE BEEN HERE
                if self.threshold_reset_method.name == 'adapt_sum_slow_fast':
                    voltage_t1 = threshold_t1
                #---------------------------------------------------------------------------------------------------------------------
                #---------below is an option you choose for reseting based on values of model at the biological spike----------------- 
                #---------or at the time where model voltage crosses model threshold--------------------------------------------------
                #---------------------------------------------------------------------------------------------------------------------
                #(voltage_t0, threshold_t0, AScurrents_t0) = self.reset(interpolated_model_spike_voltage, interpolated_model_spike_voltage, AScurrents_t1) #USE THIS IF YOU WANT TO USE USE MODEL VALUES AT MODEL SPIKE
                (voltage_t0, threshold_t0, AScurrents_t0, bad_reset_flag) = self.reset(voltage_t1, threshold_t1, AScurrents_t1)  #USE THIS IF YOU WANT TO USE MODEL VALUES AT TIME OF BIOLOGICAL SPIKE
                #---------------------------------------------------------------------------------------------------------------------
            else:
                (voltage_t0, threshold_t0, AScurrents_t0) = None, None, None
        
        return {
            'voltage': voltage_out, 
            'threshold': threshold_out, 
            'AScurrent_matrix': AScurrent_matrix, 

            'grid_model_spike_time': grid_model_spike_time,
            'interpolated_model_spike_time': interpolated_model_spike_time,
            
            'grid_model_spike_voltage': grid_model_spike_voltage,
            'interpolated_model_spike_voltage': interpolated_model_spike_voltage,

            'voltage_t0': voltage_t0, 
            'threshold_t0': threshold_t0, 
            'AScurrents_t0': AScurrents_t0, 

            'grid_bio_spike_model_voltage': voltage_t1, 
            'grid_bio_spike_model_threshold': threshold_t1            
            }

def find_first_model_spike(voltage, threshold, voltage_t1, threshold_t1, dt):
    num_time_steps = len(voltage)

    for time_step in xrange(num_time_steps): 
        if voltage[time_step] > threshold[time_step]:
            grid_model_spike_time = dt * (time_step-1)
            grid_model_spike_voltage = voltage[time_step-1]
            
            interpolated_model_spike_time = glif_neuron.interpolate_spike_time(dt, time_step-1, 
                                                                               threshold[time_step-1], threshold[time_step], 
                                                                               voltage[time_step-1], voltage[time_step])
            
            interpolated_model_spike_voltage = interpolate_spike_voltage(dt, time_step-1, 
                                                                         threshold[time_step-1], threshold[time_step], 
                                                                         voltage[time_step-1], voltage[time_step])
                
            return grid_model_spike_time, grid_model_spike_voltage, interpolated_model_spike_time, interpolated_model_spike_voltage

    # if the last voltage is above threshold and there hasn't already been a spike
    if voltage_t1 > threshold_t1: 
        grid_model_spike_time = dt * ( num_time_steps - 1 )
        grid_model_spike_voltage = voltage_t1
            
        interpolated_model_spike_time = glif_neuron.interpolate_spike_time(dt, num_time_steps - 1, threshold[num_time_steps-1], threshold_t1, voltage[num_time_steps-1], voltage_t1)
        interpolated_model_spike_voltage = interpolate_spike_voltage(dt, num_time_steps, threshold[-1], threshold_t1, voltage[-1], voltage_t1)

        return grid_model_spike_time, grid_model_spike_voltage, interpolated_model_spike_time, interpolated_model_spike_voltage


    return None, None, None, None

def extrapolate_model_spike_from_endpoints(neuron, voltage, threshold, voltage_t1, threshold_t1, dt):
    
    #--extrapolate using first point in ISI and last point in ISI
    num_time_steps = len(voltage)
    
    interpolated_model_spike_time = extrapolate_spike_time(dt, num_time_steps, threshold[0], threshold_t1, voltage[0], voltage_t1)
    interpolated_model_spike_voltage = extrapolate_spike_voltage(dt, num_time_steps, threshold[0], threshold_t1, voltage[0], voltage_t1)
            
    grid_model_spike_time = np.ceil(interpolated_model_spike_time / dt) * dt  # grid spike time based off extrapolated spike time
    grid_model_spike_voltage = interpolated_model_spike_voltage

    result = grid_model_spike_time, grid_model_spike_voltage, interpolated_model_spike_time, interpolated_model_spike_voltage
    

    return result

    

def extrapolate_model_spike_from_endpoints_single_tau(neuron, voltage, threshold, voltage_t1, threshold_t1, dt):
    tau_m = neuron.tau_m
    num_time_steps = len(voltage)
    ii = np.floor(tau_m/dt)
    starting_ind = max(0,(num_time_steps - ii)) 
    result = extrapolate_model_spike_from_endpoints(neuron, voltage[starting_ind:], threshold[starting_ind:], voltage_t1, threshold_t1, dt)
    
    return result
    
def extrapolate_spike_time(dt, num_time_steps, threshold_t0, threshold_t1, voltage_t0, voltage_t1):
    """ Given two voltage and threshold values and an interval between them, extrapolate a spike time
    by intersecting lines the thresholds and voltages. """
    return glif_neuron.line_crossing_x(dt * num_time_steps, voltage_t0, voltage_t1, threshold_t0, threshold_t1)

def extrapolate_spike_voltage(dt, num_time_steps, threshold_t0, threshold_t1, voltage_t0, voltage_t1):
    """ Given two voltage and threshold values and an interval between them, extrapolate a spike time
    by intersecting lines the thresholds and voltages. """
    return glif_neuron.line_crossing_y(dt * num_time_steps, voltage_t0, voltage_t1, threshold_t0, threshold_t1)
    
def interpolate_spike_voltage(dt, time_step, threshold_t0, threshold_t1, voltage_t0, voltage_t1):
    """ Given two voltage and threshold values, the dt between them and the initial time step, interpolate
    a spike time within the dt interval by intersecting the two lines. """
    return time_step*dt + glif_neuron.line_crossing_y(dt, voltage_t0, voltage_t1, threshold_t0, threshold_t1)



#!/usr/bin/python
import sys, logging
import os
import json
import shutil
import argparse
import copy
import numpy as np
import shutil

from allensdk.config.manifest import Manifest

import allensdk.internal.core.lims_utilities as lims_utilities
import allensdk.core.json_utilities as json_utilities

from allensdk.internal.ephys.core_feature_extract import *
from allensdk.ephys.ephys_features import FeatureError


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('input_json')
    parser.add_argument('output_json')
    parser.add_argument('--log_level')
    parser.add_argument('--output_directory')
    
    args = parser.parse_args()

    if args.log_level:
        logging.getLogger().setLevel(args.log_level)

    return args

def main():
    args = parse_args()

    input_data = json_utilities.read(args.input_json)
    input_err, stim_types = input_data

    output_data = copy.deepcopy(input_err)

    err_wkfs = output_data['well_known_files']
    nwb_file = lims_utilities.get_well_known_file_by_type(err_wkfs, lims_utilities.NWB_FILE_TYPE_ID)
    storage_directory = args.output_directory or output_data['storage_directory']
    # move code to help make data extraction compatible with ephys qc tool
    try:
        sweep_list, sweep_features = extract_data(output_data, nwb_file)
    except FeatureError as e:
        logging.error("Error computing cell features, auto-failing cell: %s" % e.message)
        output_data["workflow_state"] = "auto_failed"
        json_utilities.write(args.output_json, output_data)
        return
    #
    
    # embed spike times in NWB file
    logging.debug("Embedding spike times")
    tmp_nwb_file = os.path.join(storage_directory, os.path.basename(nwb_file) + '.tmp')
    out_nwb_file = os.path.join(storage_directory, os.path.basename(nwb_file))

    shutil.copy(nwb_file, tmp_nwb_file)
    for sweep in sweep_list:
        sweep_num = sweep['sweep_number']

        if sweep_num not in sweep_features:
            continue

        try:
            spikes = sweep_features[sweep_num]['spikes']
            spike_times = [ s['threshold_t'] for s in spikes ]
            NwbDataSet(tmp_nwb_file).set_spike_times(sweep_num, spike_times)
        except Exception as e:
            logging.info("sweep %d has no sweep features. %s", sweep_num, e.message)

    try:
        shutil.move(tmp_nwb_file, out_nwb_file)
    except OSError as e:
        logging.error("Problem renaming file: %s -> %s" % (tmp_nwb_file, out_nwb_file))
        raise e

    qc_fig_dir = os.path.join(storage_directory, 'qc_figures')
    save_qc_figures(qc_fig_dir, nwb_file, output_data, True)

    # regenerating this file
    features_json = os.path.join(storage_directory, "%d_ephys_features.json" % output_data['id'])
    json_utilities.write(features_json, output_data)
    lims_utilities.append_well_known_file(output_data['well_known_files'], features_json)

    # write output json files
    json_utilities.write(args.output_json, output_data)

            
if __name__ == "__main__": 
    main()



import json, os
import sys
import subprocess

import run_observatory_thumbnails as robsth
import allensdk.internal.core.lims_utilities as lu
from allensdk.internal.core.lims_pipeline_module import run_module, PipelineModule
import allensdk.core.json_utilities as ju
from allensdk.config.manifest import Manifest

def get_container_info(container_id):
    res = lu.query("""
select * from ophys_experiments oe
where experiment_container_id = %d
and oe.workflow_state != 'failed'
""" % container_id)
    return res

def debug(container_id, local=False, plots=None):
    SCRIPT = "/data/informatics/CAM/analysis/allensdk/allensdk/internal/pipeline_modules/run_observatory_container_thumbnails.py"
    SDK_PATH = "/data/informatics/CAM/analysis/allensdk/"
    OUTPUT_DIR = "/data/informatics/CAM/analysis/containers"

    container_dir = os.path.join(OUTPUT_DIR, str(container_id))

    input_data = []
    for exp in get_container_info(container_id):
        exp_data = robsth.get_input_data(exp['id'])
        exp_input_json = os.path.join(exp_data["output_directory"], "input.json")
        input_data.append(dict(
                input_json=exp_input_json,
                output_json=os.path.join(exp_data["output_directory"], "output.json")
                ))

        Manifest.safe_make_parent_dirs(exp_input_json)
        ju.write(exp_input_json, exp_data)

    run_module(SCRIPT,
               input_data,
               container_dir,
               sdk_path=SDK_PATH,
               pbs=dict(vmem=32,
                        job_name="cthumbs_%d"% container_id,
                        walltime="10:00:00"),
               local=local,
               optional_args=['--types='+','.join(plots)] if plots else None)

def main():
    mod = PipelineModule()
    mod.parser.add_argument("--types", default=','.join(robsth.PLOT_TYPES))
    mod.parser.add_argument("--threads", default=4)
    
    data = mod.input_data()
    types = mod.args.types.split(',')

    for input_file in data:
        exp_input_json = input_file['input_json']
        exp_output_json = input_file['output_json']

        exp_input_data = ju.read(exp_input_json)

        nwb_file, analysis_file, output_directory = robsth.parse_input(exp_input_data)
        
        robsth.build_experiment_thumbnails(nwb_file, analysis_file, output_directory, 
                                           types, mod.args.threads)

if __name__=='__main__': main()



import logging
from allensdk.internal.core.lims_pipeline_module import (PipelineModule,
                                                         run_module)
import allensdk.core.json_utilities as ju
import allensdk.internal.core.lims_utilities as lu
from allensdk.internal.brain_observatory import ophys_session_decomposition as osd
from multiprocessing import Pool
import os

DEBUG_CHANNELS = ["data", "piezo"]
DEBUG_WIDTH = 512
DEBUG_HEIGHT = 256
DEBUG_ITEMSIZE = 2
DEBUG_N_PLANES = 6

def create_fake_metadata(exp_dir, raw_path, channels=None,
                         width=DEBUG_WIDTH, height=DEBUG_HEIGHT,
                         itemsize=DEBUG_ITEMSIZE, n_planes=DEBUG_N_PLANES):
    metadata = []
    size = os.stat(raw_path).st_size
    if channels is None:
        channels = DEBUG_CHANNELS
    n_frames = size/(itemsize*width*height)
    frames_per_plane = n_frames/n_planes/len(channels)
    for plane in range(n_planes):
        experiment_id = plane
        outfile = os.path.join(exp_dir, "plane_{}.h5".format(plane))
        frame_meta = []
        for i, channel in enumerate(channels):
            byte_offset = width * height * itemsize * \
                          (plane * len(channels) + i)
            strides = [width*height*itemsize*n_planes*len(channels),
                       width*itemsize,
                       itemsize]
            frame_meta.append({"byte_offset": byte_offset,
                               "channel": i+1,
                               "channel_description": channel,
                               "frame_description": "plane_{}".format(plane),
                               "dtype": ">u{}".format(itemsize),
                               "position_offset": [None, 0, 0],
                               "shape": [frames_per_plane, height, width],
                               "strides": strides})
        metadata.append({"output_file": outfile,
                         "experiment_id": experiment_id,
                         "frame_metadata": frame_meta})
    return metadata


def debug(experiment_id, local=False, raw_path=None):
    OUTPUT_DIRECTORY = "/data/informatics/CAM/ophys_decomp"
    SDK_PATH = "/data/informatics/CAM/ophys_decomp/allensdk"
    SCRIPT = ("/data/informatics/CAM/ophys_decomp/allensdk/allensdk/"
              "internal/pipeline_modules/run_ophys_session_decomposition.py")

    exp_dir = os.path.join(OUTPUT_DIRECTORY, str(experiment_id))

    if raw_path is not None:
        conversion_definitions = create_fake_metadata(exp_dir, raw_path)
        input_data = {"raw_filename": raw_path,
                      "frame_metadata": conversion_definitions}
    else:
        raise NotImplementedError("No real examples exist yet")

    run_module(SCRIPT,
               input_data,
               exp_dir,
               sdk_path=SDK_PATH,
               pbs=dict(vmem=160,
                        job_name="ophys_decomp_%d"% experiment_id,
                        walltime="36:00:00"),
               local=local)


def convert_frame(conversion_definition):
    raw_filename = conversion_definition["input_file"]
    ophys_hdf5_filename = conversion_definition["data_output_file"]
    auxiliary_hdf5_filename = conversion_definition["auxiliary_output_file"]
    experiment_id = conversion_definition["experiment_id"]
    frame_metadata = conversion_definition["frame_metadata"]
    osd.export_frame_to_hdf5(raw_filename, ophys_hdf5_filename,
                             auxiliary_hdf5_filename, frame_metadata)
    return experiment_id, ophys_hdf5_filename, auxiliary_hdf5_filename


def parse_input(data):
    '''Load all input data from the input json.'''
    conversion_definitions = data["frame_metadata"]
    for item in conversion_definitions:
        item["input_file"] = data["raw_filename"]
    return conversion_definitions


def main():
    mod = PipelineModule("Decompose ophys session into individual planes.")
    mod.parser.add_argument("-t", "--threads", type=int, default=4)

    input_data = mod.input_data()
    conversion_definitions = parse_input(input_data)

    if mod.args.threads > 1:
        pool = Pool(processes=mod.args.threads)
        output = pool.map(convert_frame, conversion_definitions)
    else:
        output= []
        for definition in conversion_definitions:
            output.append(convert_frame(definition))

    output_data = {}
    for eid, ophys_file, auxiliary_file in output:
        output_data[eid] = {"ophys_data": ophys_file,
                            "auxiliary_data": auxiliary_file}

    mod.write_output_data(output_data)

if __name__ == "__main__": main()



import logging
import sys
import numpy as np
import h5py

from qc_support import *
from lab_notebook_reader import *

from allensdk.internal.core.lims_pipeline_module import PipelineModule
from allensdk.core.nwb_data_set import NwbDataSet


# manual keys are values that can be passed in through input.json.
# these values are used if the particular value cannot be computed.
# a better name might be 'DEFAULT_VALUE_KEYS'
MANUAL_KEYS = ['manual_seal_gohm', 'manual_initial_access_resistance_mohm', 'manual_initial_input_mohm' ]

# names of blocks used in output.json
# for sweep-specific data:
JSON_BLOCK_SWEEP_DATA = "sweep_data"
# for data that applies to the entire experiment:
JSON_BLOCK_EXPERIMENT_DATA = "experiment_data"

########################################################################
# bootstrapping code
# this module doesn't know anything about what's in the supplied NWB 
#   file and simply assumes that it's an IVSCC file. it must find and
#   fetch data as appropriate
# processing requires being able to pull out sweeps of specific types.
#   create an index of those types here, and provide accessor functions
#   for the indexed data


# local globals used to avoid having to pass parameters to functions that
#   can be several calls deep before they're needed
# consider refactoring into QC class to avoid this approach
sweep_stim_map = None
stim_sweep_map = None
sweep_list = None
nwb_file_name = None

# reads the NWB file and generates a mapping between sweep name and 
#   stimulus code, and vice versa
def build_sweep_stim_map():
    global sweep_stim_map, stim_sweep_map, nwb_file_name, sweep_list
    try:
        nwb_file = h5py.File(nwb_file_name, "r")
    except:
        raise Exception ("Unable to open input NWB file '%s'" % str(nwb_file_name))
    print("Opened '%s'" % str(nwb_file_name))
    sweep_stim_map = {}
    stim_sweep_map = {}
    sweep_list = []
    acq = nwb_file["acquisition/timeseries"]
    for sweep in acq:
        # if string storage is variable length, data appears to be stored
        #   or retrieved as an array of strings, so we need to take the
        #   first element. this happens with Igor-generated files.
        # if the string is stored with fixed length, data appears to be
        #   stored as a string, so we must take the entire value
        stim = acq[sweep]["aibs_stimulus_description"].value[0]
        if len(stim) == 1:
            stim = acq[sweep]["aibs_stimulus_description"].value
        stim_sweep_map[stim] = sweep
        #print "%s (%s) : %s (%s)" % (sweep, type(sweep), stim, type(stim))
        sweep_stim_map[sweep] = stim
        sweep_list.append(sweep)
    nwb_file.close()

# fetches stimulus code for a given sweep name, or None if no stimulus
#   was found for the specified sweep
def get_sweep_name_by_stimulus_code(stim_name):
    """ Returns the first sweep name that uses the specified stimulus
        type. 'First' does not mean lowest sweep number, only the first
        one found using a [random] dictionary search.

        Input: stimulus name (string)

        Output: sweep name (string), or None if no sweep found for this stim
    """
    global sweep_stim_map
    for k,v in stim_sweep_map.iteritems():
        if k.startswith(stim_name):
            return v
    return None

        
# returns True if stimulus name for specified sweep indicates the sweep
#   is a ramp and False otherwise
def sweep_is_ramp(sweep_name):
    """ Input: sweep name (string)

        Output: boolean (True if sweep is ramp, False otherwise)
    """
    global sweep_stim_map
    return sweep_stim_map[sweep_name].startswith('C1RP')


# old code based on using NwbDataSet objects. provide a way to
#   create them in order to leverage old code as much as possible
def get_sweep_data(sweep_name):
    """ Input: sweep name (string)
        
        Output: NwbDataSet object
    """
    global nwb_file_name
    try:
        num = int(sweep_name.split('_')[-1])
    except:
        print("Unable to parse sweep number from '%s'" % str(sweep_name))
        raise
    return NwbDataSet(nwb_file_name).get_sweep(num)


# functions to lookup a sweep having the desired stimulus code
# NOTE: if multiple instance exist then only one instance is returned
def get_blowout_sweep():
    """ Returns NwbDataSet for the blowout sweep, or None if it's absent
    """
    sweep_name = get_sweep_name_by_stimulus_code('EXTPBLWOUT')
    if sweep_name is None:
        return None
    return get_sweep_data(sweep_name)

def get_bath_sweep():
    """ Returns NwbDataSet for the bath sweep, or None if it's absent
    """
    sweep_name = get_sweep_name_by_stimulus_code('EXTPINBATH')
    if sweep_name is None:
        return None
    return get_sweep_data(sweep_name)

def get_seal_sweep():
    """ Returns NwbDataSet for the seal sweep, or None if it's absent
    """
    sweep_name = get_sweep_name_by_stimulus_code('EXTPCllATT')
    if sweep_name is None:
        return None
    return get_sweep_data(sweep_name)

def get_breakin_sweep():
    """ Returns NwbDataSet for the breakin sweep, or None if it's absent
    """
    sweep_name = get_sweep_name_by_stimulus_code('EXTPBREAKN')
    if sweep_name is None:
        return None
    return get_sweep_data(sweep_name)


########################################################################


########################################################################
# QC-relevant feature extraction code

# cell-level values (for ephys_roi_results) 
def cell_level_features(jin, jout, sweep_tag_list, manual_values):
    """
    """
    output_data = {}
    jout[JSON_BLOCK_EXPERIMENT_DATA] = output_data
    # measure blowout voltage
    try:
        blowout_data = get_blowout_sweep()
        blowout = measure_blowout(blowout_data['response'], 
                                  blowout_data['index_range'][0])
        output_data['blowout_mv'] = blowout
    except:
        msg = "Blowout is not available"
        sweep_tag_list.append(msg)
        logging.warning(msg)
        output_data['blowout_mv'] = None


    # measure "electrode 0"
    try:
        bath_data = get_bath_sweep()
        e0 = measure_electrode_0(bath_data['response'], 
                                 bath_data['sampling_rate'])
        output_data['electrode_0_pa'] = e0
    except:
        msg = "Electrode 0 is not available"
        sweep_tag_list.append(msg)
        logging.warning(msg)
        output_data['electrode_0_pa'] = None


    # measure clamp seal
    try:
        seal_data = get_seal_sweep()
        seal = measure_seal(seal_data['stimulus'], 
                            seal_data['response'], 
                            seal_data['sampling_rate'])
        # error may arise in computing seal, which falls through to
        #   exception handler. if seal computation didn't fail but
        #   computation generated invalid value, trigger same 
        #   exception handler with different error
        if seal is None or not np.isfinite(seal):
            raise Exception("Could not compute seal")
    except:
        # seal is not available, for whatever reason. log error
        msg = "Seal is not available"
        sweep_tag_list.append(msg)
        logging.warning(msg)
        # look for manual seal value and use it if it's available
        seal = manual_values.get('manual_seal_gohm', None)
        if seal is not None:
            logging.info("using manual seal value: %f" % seal)
            sweep_tag_list.append("Seal set using manual value")
    output_data["seal_gohm"] = seal


    # measure input and series resistance
    # this requires two steps -- finding the breakin sweep, and then 
    #   analyzing it
    # if the value is unavailable then check to see if it was set manually
    breakin_data = None
    try:
        breakin_data = get_breakin_sweep()
    except:
        logging.warning("Error reading breakin sweep.")
        sweep_tag_list.append("Breakin sweep not found")

    ir = None   # input resistance
    sr = None   # series resistance
    if breakin_data is not None:
        ###########################
        # input resistance
        try:
            ir = measure_input_resistance(breakin_data['stimulus'], 
                                          breakin_data['response'], 
                                          breakin_data['sampling_rate'])
        except:
            logging.warning("Error reading input resistance.")
        # apply manual value if it's available
        if ir is None:
            sweep_tag_list.append("Input resistance is not available")
            ir = manual_values.get('manual_initial_input_mohm', None)
            if ir is not None:
                msg = "Using manual value for input resistance"
                logging.info(msg)
                sweep_tag_list.append(msg);
        ###########################
        # initial access resistance
        try:
            sr = measure_initial_access_resistance(breakin_data['stimulus'], 
                                               breakin_data['response'], 
                                               breakin_data['sampling_rate'])
        except:
            logging.warning("Error reading initial access resistance.")
        # apply manual value if it's available
        if sr is None:
            sweep_tag_list.append("Initial access resistance is not available")
            sr = manual_values.get('manual_initial_access_resistance_mohm', None)
            if sr is not None:
                msg = "Using manual initial access resistance"
                logging.info(msg)
                sweep_tag_list.append(msg)
    #
    output_data['input_resistance_mohm'] = ir
    output_data["initial_access_resistance_mohm"] = sr

    sr_ratio = None # input access resistance ratio
    if ir is not None and sr is not None:
        try:
            sr_ratio = sr / ir
        except:
            pass    # let sr_ratio stay as None
    output_data['input_access_resistance_ratio'] = sr_ratio


##############################
def sweep_level_features(jin, jout, sweep_tag_list):
    """
    """
    global sweep_list
    # pull out features from each sweep (for ephys_sweeps)
    cnt = 0
    jout[JSON_BLOCK_SWEEP_DATA] = {}
    for sweep_name in sweep_list:
        # pull data streams from file
        sweep_num = int(sweep_name.split('_')[-1])
        try:
            sweep_data = NwbDataSet(nwb_file_name).get_sweep(sweep_num)
        except:
            logging.warning("Error reading sweep %d" % sweep_num)
            continue
        sweep = {}
        jout[JSON_BLOCK_SWEEP_DATA][sweep_name] = sweep

        # don't process voltage clamp sweeps
        if sweep_data["stimulus_unit"] == "Volts":
            continue    # voltage-clamp

        volts = sweep_data['response']
        current = sweep_data['stimulus']
        hz = sweep_data['sampling_rate']
        idx_start, idx_stop = sweep_data['index_range']

        # measure Vm and noise before stimulus
        idx0, idx1 = get_first_vm_noise_epoch(idx_start, current, hz)
        _, rms0 = measure_vm(1e3 * volts[idx0:idx1])

        sweep["pre_noise_rms_mv"] = float(rms0)

        # measure Vm and noise at end of recording
        # only do so if acquisition not truncated 
        # do not check for ramps, because they do not have enough time to recover
        mean1 = None
        sweep_not_truncated = ( idx_stop == len(current) - 1 )
        if sweep_not_truncated and not sweep_is_ramp(sweep_name):
            idx0, idx1 = get_last_vm_epoch(idx_stop, current, hz)
            mean1, _ = measure_vm(1e3 * volts[idx0:idx1])
            idx0, idx1 = get_last_vm_noise_epoch(idx_stop, current, hz)
            _, rms1 = measure_vm(1e3 * volts[idx0:idx1])
            sweep["post_vm_mv"] = float(mean1)
            sweep["post_noise_rms_mv"] = float(rms1)

        # measure Vm and noise over extended interval, to check stability
        stim_start = find_stim_start(idx_start, current)
        sweep['stimulus_start_time'] = stim_start / sweep_data['sampling_rate']

        idx0, idx1 = get_stability_vm_epoch(idx_start, stim_start, hz)
        mean2, rms2 = measure_vm(1000 * volts[idx0:idx1])

        slow_noise = float(rms2)
        sweep["slow_vm_mv"] = float(mean2)
        sweep["slow_noise_rms_mv"] = float(rms2)

        # for now (mid-feb 15), make vm_mv the same for pre and slow
        mean0 = mean2
        sweep["pre_vm_mv"] = float(mean0)
        if mean1 is not None:
            delta = abs(mean0 - mean1)
            sweep["vm_delta_mv"] = float(delta)
        else:
            # Use None as 'nan' still breaks the ruby strategies
            sweep["vm_delta_mv"] = None

        # compute stimulus duration, amplitude, interal
        stim_amp, stim_dur = find_stim_amplitude_and_duration(idx_start, current, hz)
        stim_int = find_stim_interval(idx_start, current, hz)

        sweep['stimulus_amplitude'] = stim_amp * 1e12
        sweep['stimulus_duration'] = stim_dur
        sweep['stimulus_interval'] = stim_int

        tag_list = []
        for i in range(len(sweep_tag_list)):
            tag = {}
            tag["name"] = sweep_tag_list[i]
            tag_list.append(tag)
        sweep["ephys_sweep_tags"] = tag_list


# create a summary table of sweeps and stimuli
def summarize_sweeps(jin, jout):
    global nwb_file_name
    # build stimulus name map
    stim_type_name_map = {}
    for group_name, raw_names in jin["ephys_raw_stimulus_names"].iteritems():
        for n in raw_names:
            stim_type_name_map[n] = group_name

    h5_file_name = jin.get("input_h5", None)
    notebook = create_lab_notebook_reader(nwb_file_name, h5_file_name)
    borg = h5py.File(nwb_file_name, 'r')

    # two json blocks to store data in
    exp_data = jout[JSON_BLOCK_EXPERIMENT_DATA]
    swp_data = jout[JSON_BLOCK_SWEEP_DATA]
    #jout["sweep_summary"] = output_data

#    # verify input file generated by Igor
#    generated_by = borg["general/generated_by"].value
#    igor = False
#    for row in generated_by:
#        if row[0] == "Program" and row[1].startswith('Igor'):
#            igor = True
#            break
#    if not igor:
#        print("Error -- File not recognized as Igor-generated NWB file")
#        return -1

    # validated nwb files can have different types of string storage
    # problem seems to be related to h5py and if string is stored as
    #   fixed- or variable-width. assume that string is more than one
    #   character and try to auto-correct for this issue
    session_date = borg["session_start_time"].value
    if len(session_date) == 1:
        session_date = session_date[0]
    exp_data['recording_date'] = session_date

    # get sampling rate
    # use same output strategy as h5-nwb converter
    # pick the sampling rate from the first iclamp sweep
    # TODO: figure this out for multipatch
    sampling_rate = None
    for sweep_name in borg["acquisition/timeseries"]:
        sweep_ts = borg["acquisition/timeseries"][sweep_name]
        ancestry = sweep_ts.attrs["ancestry"]
        if "CurrentClamp" in ancestry[-1]:
            if sampling_rate is None:
                sampling_rate = sweep_ts["starting_time"].attrs["rate"]
                break
    if sampling_rate is None:
        raise Exception("Unable to determine sampling rate from current clamp sweep.")
    exp_data['sampling_rate'] = sampling_rate
#    sweep_data = []
#    output_data["sweep_summary"] = sweep_data

    # read sweep-specific data
    for sweep_name in borg["acquisition/timeseries"]:
        # get h5 timeseries object, and the sweep number
        sweep_ts = borg["acquisition/timeseries"][sweep_name]
        sweep_num = int(sweep_name.split('_')[-1])
        #sweep_num = int(sweep_name[:-4].split('_')[-1]) # for reading igor nwb
        # fetch stim name from lab notebook
        stim_name = notebook.get_value("Stim Wave Name", sweep_num, "")
        if len(stim_name) == 0:
            raise Exception("Could not read stimulus wave name from lab notebook for sweep %d" % sweep_num)

        # stim units are based on timeseries type
        ancestry = sweep_ts.attrs["ancestry"]
        if "CurrentClamp" in ancestry[-1]:
            stim_units = 'pA'
        elif "VoltageClamp" in ancestry[-1]:
            stim_units = 'mV'
        else:
            # it's probably OK to skip this sweep and put a 'continue' 
            #   here instead of an exception, but wait until there's
            #   an actual error and investigate the data before doing so
            raise Exception("Unable to determine clamp mode in " + sweep_name)

        # stim name stored in database as, eg, C2SSTRIPLE150429
        # stim name in igor nwb stored as C2SSTRIPLE150429_DA_0
        # -> need to strip last 5 chars off to make match for lookup
        stim_type_name = stim_type_name_map.get(stim_name[:-5], None)
        if stim_type_name is None:
            raise Exception("Could not find stimulus raw name (\"%s\") for sweep %d." % (stim_name, sweep_num))

        # voltage-clamp sweeps shouldn't have a record yet -- make one
        if sweep_name not in swp_data:  
            swp_data[sweep_name] = {}
        info = swp_data[sweep_name]

        # sweep number
        info["sweep_number"] = sweep_num
        # bridge balance
        bridge_balance = notebook.get_value("Bridge Bal Value", sweep_num, None)
        # IT-14677
        # if bridge_balance is None, that's OK. do NOT change it to NaN

        info["bridge_balance_mohm"] = bridge_balance
        # stimulus units
        info["stimulus_units"] = stim_units
        # leak_pa (bias current)
        bias_current = notebook.get_value("I-Clamp Holding Level", sweep_num, None)
        # IT-14677
        # if bias_current is None, that's OK. do NOT change it to NaN

        info["leak_pa"] = bias_current
        #
        # ephys stim info
        scale_factor = notebook.get_value("Scale Factor", sweep_num, None)
        if scale_factor is None:
            raise Exception("Unable to read scale factor for " + sweep_name)
        # PBS-229 change stim name by appending set_sweep_count
        cnt = notebook.get_value("Set Sweep Count", sweep_num, 0)
        stim_name_ext = stim_name.split('_')[0] + "[%d]" % int(cnt)
        info["ephys_stimulus"] = {
            #'description': stim_name,
            'description': stim_name_ext,
            'amplitude': scale_factor,
            'ephys_stimulus_type': { 'name': stim_type_name }
        }
    #
    borg.close()


########################################################################
########################################################################


def main(jin):
    # to avoid passing arguments amongs the many functions and procedures,
    #   set a global value 'nwb_file_name' that each function can read
    global nwb_file_name
    nwb_file_name = jin["input_nwb"]

    # initialize index of stimuli and sweeps
    build_sweep_stim_map()

    # TODO Document manual keys, and what they're for
    manual_values = {}
    for k in MANUAL_KEYS:
        if k in jin:
            manual_values[k] = jin[k]

    # dictionary for json output
    jout = {}

    # list of messages (tags) that log information about this sweep set
    #  (eg, 'Seal not available')
    sweep_tag_list = []

    cell_level_features(jin, jout, sweep_tag_list, manual_values)
    # sweep level data. first pull out QC-relevant metrics, then store
    #   stimulus info with that data
    sweep_level_features(jin, jout, sweep_tag_list)
    summarize_sweeps(jin, jout)

    return jout



if __name__ == "__main__":
    # read module input. PipelineModule object automatically parses the 
    #   command line to pull out input.json and output.json file names
    module = PipelineModule()
    jin = module.input_data()   # loads input.json
    jout = main(jin)
    module.write_output_data(jout)  # writes output.json



import os
from allensdk.internal.core.lims_pipeline_module import (PipelineModule,
                                                         run_module)
from allensdk.internal.brain_observatory import (eye_calibration,
                                                 itracker_utils)
import allensdk.internal.core.lims_utilities as lu
import numpy as np
import h5py

EYE_RADIUS = 0.1682
CM_PER_PIXEL = 10.2/10000

def get_wkf(wkf_type, experiment_id):
    wkf = lu.query("""
select CONCAT(wkf.storage_directory, wkf.filename) as path
from well_known_files wkf
join well_known_file_types wkft on wkft.id = wkf.well_known_file_type_id
where wkft.name LIKE '{}' and wkf.attachable_id = {}
""".format(wkf_type, experiment_id))[0]["path"]
    return wkf

def debug(experiment_id, local=False):
    OUTPUT_DIRECTORY = "/data/informatics/CAM/eye_calibration"
    SDK_PATH = "/data/informatics/CAM/eye_calibration/allensdk"
    SCRIPT = ("/data/informatics/CAM/eye_calibration/allensdk/allensdk"
              "/internal/pipeline_modules/run_ophys_eye_calibration.py")

    frame_width = 640
    frame_height = 480

    cr_file = get_wkf("EyeTracking Corneal Reflection", experiment_id)
    pupil_file = get_wkf("EyeTracking Pupil", experiment_id)

    exp_info = lu.query("""
select *
from ophys_sessions os
where os.id = {}
""".format(experiment_id))[0]

    exp_dir = os.path.join(OUTPUT_DIRECTORY, str(experiment_id))
    # clear out missing values to let us get defaults
    for key, value in list(exp_info.iteritems()):
        if value is None:
            del exp_info[key]

    input_data = {
        "cr_params_file": cr_file,
        "pupil_params_file": pupil_file,
        "frame_width": frame_width,
        "frame_height": frame_height,
        "output_file": os.path.join(exp_dir,
                                    "eye_tracking_to_screen_mapping.h5"),
        "monitor_position_x_mm": exp_info.get(
            "screen_center_x_mm", eye_calibration.MONITOR_POSITION_OLD[0]*10),
        "monitor_position_y_mm": exp_info.get(
            "screen_center_y_mm", eye_calibration.MONITOR_POSITION_OLD[1]*10),
        "monitor_position_z_mm": exp_info.get(
            "screen_center_z_mm", eye_calibration.MONITOR_POSITION_OLD[2]*10),
        "monitor_rotation_x_deg": exp_info.get("screen_rotation_x_deg", 0),
        "monitor_rotation_y_deg": exp_info.get("screen_rotation_y_deg", 0),
        "monitor_rotation_z_deg": exp_info.get("screen_rotation_z_deg", 0),
        "camera_position_x_mm": exp_info.get(
            "camera_center_x_mm", eye_calibration.CAMERA_POSITION_OLD[0]*10),
        "camera_position_y_mm": exp_info.get(
            "camera_center_y_mm", eye_calibration.CAMERA_POSITION_OLD[1]*10),
        "camera_position_z_mm": exp_info.get(
            "camera_center_z_mm", eye_calibration.CAMERA_POSITION_OLD[2]*10),
        "camera_rotation_x_deg": exp_info.get(
            "camera_rotation_x_deg",
            eye_calibration.CAMERA_ROTATIONS_OLD[0]*180/np.pi),
        "camera_rotation_y_deg": exp_info.get(
            "camera_rotation_y_deg",
            eye_calibration.CAMERA_ROTATIONS_OLD[1]*180/np.pi),
        "camera_rotation_z_deg": exp_info.get(
            "camera_rotation_z_deg",
            eye_calibration.CAMERA_ROTATIONS_OLD[2]*180/np.pi),
        "led_position_x_mm": exp_info.get(
            "led_center_x_mm", eye_calibration.LED_POSITION_ORIGINAL[0]*10),
        "led_position_y_mm": exp_info.get(
            "led_center_y_mm", eye_calibration.LED_POSITION_ORIGINAL[1]*10),
        "led_position_z_mm": exp_info.get(
            "led_center_z_mm", eye_calibration.LED_POSITION_ORIGINAL[2]*10)
    }

    # TEMPORARY HACKS TO DEAL WITH BAD DATA IN LIMS
    # TODO: REMOVE WHEN DATAFIXES DONE
    if input_data["monitor_position_x_mm"] == -86.2:
        input_data["monitor_position_x_mm"] = \
            eye_calibration.MONITOR_POSITION_NEW[0]*10
        input_data["monitor_position_y_mm"] = \
            eye_calibration.MONITOR_POSITION_NEW[1]*10
        input_data["monitor_position_z_mm"] = \
            eye_calibration.MONITOR_POSITION_NEW[2]*10

    run_module(SCRIPT,
               input_data,
               exp_dir,
               sdk_path=SDK_PATH,
               local=local)


def parse_input_data(data):
    cr_params = np.load(data["cr_params_file"])
    pupil_params = np.load(data["pupil_params_file"])
    frame_width = data["frame_width"]
    frame_height = data["frame_height"]
    output_file = data["output_file"]
    monitor_position = np.array([
        float(data['monitor_position_x_mm'])/10.0,
        float(data['monitor_position_y_mm'])/10.0,
        float(data['monitor_position_z_mm'])/10.0,
    ])
    monitor_rotations = np.array([
        float(data['monitor_rotation_x_deg'])*np.pi/180,
        float(data['monitor_rotation_y_deg'])*np.pi/180,
        float(data['monitor_rotation_z_deg'])*np.pi/180,
    ])
    camera_position = np.array([
        float(data['camera_position_x_mm'])/10.0,
        float(data['camera_position_y_mm'])/10.0,
        float(data['camera_position_z_mm'])/10.0,
    ])
    camera_rotations = np.array([
        float(data['camera_rotation_x_deg'])*np.pi/180,
        float(data['camera_rotation_y_deg'])*np.pi/180,
        float(data['camera_rotation_z_deg'])*np.pi/180,
    ])
    led_position = np.array([
        float(data['led_position_x_mm'])/10.0,
        float(data['led_position_y_mm'])/10.0,
        float(data['led_position_z_mm'])/10.0,
    ])
    calibrator = eye_calibration.EyeCalibration(
        monitor_position=monitor_position,
        monitor_rotations=monitor_rotations,
        led_position=led_position,
        camera_position=camera_position,
        camera_rotations=camera_rotations,
        eye_radius=EYE_RADIUS,
        cm_per_pixel=CM_PER_PIXEL)

    cr_params, _ = itracker_utils.post_process_cr(cr_params)
    pupil_params = itracker_utils.post_process_pupil(pupil_params)
    cr_params = itracker_utils.filter_bad_params(cr_params, frame_width,
                                                 frame_height)
    pupil_params = itracker_utils.filter_bad_params(pupil_params, frame_width,
                                                    frame_height)
    return calibrator, cr_params, pupil_params, output_file


def write_output(filename, position_degrees, position_cm, areas):
    with h5py.File(filename, "w") as f:
        f.create_dataset("screen_coordinates", data=position_cm)
        f.create_dataset("screen_coordinates_spherical",
                         data=position_degrees)
        f.create_dataset("pupil_areas", data=areas)


def main():
    mod = PipelineModule()
    data = mod.input_data()
    calibrator, cr_params, pupil_params, outfile = parse_input_data(data)

    pupil_areas = calibrator.compute_area(pupil_params)
    pupil_on_monitor_deg = calibrator.pupil_position_on_monitor_in_degrees(
        pupil_params, cr_params)
    pupil_on_monitor_cm = calibrator.pupil_position_on_monitor_in_cm(
        pupil_params, cr_params)
    missing_index = np.isnan(pupil_areas) | np.isnan(pupil_on_monitor_deg.T[0])
    pupil_areas[missing_index] = np.nan
    pupil_on_monitor_deg[missing_index,:] = np.nan
    pupil_on_monitor_cm[missing_index,:] = np.nan
    write_output(outfile, pupil_on_monitor_deg, pupil_on_monitor_cm,
                 pupil_areas)

    mod.write_output_data({"screen_mapping_file": outfile})

if __name__ == "__main__": main()



import os
import sys
import allensdk.internal.model.biophysical.ephys_utils as ephys_utils
import check_fi_shift
import pandas as pd
import numpy as np
from collections import Counter
import subprocess

from allensdk.ephys.ephys_extractor \
    import EphysSweepFeatureExtractor, EphysSweepSetFeatureExtractor
import allensdk.core.json_utilities as ju
from allensdk.core.nwb_data_set import NwbDataSet
import allensdk.internal.model.biophysical.optimize as optimize
import logging

SEEDS = [1234, 1001, 4321, 1024, 2048]
FIT_BASE_DIR = os.path.join(os.path.dirname(__file__), "fits")
APICAL_DENDRITE_TYPE = 4
MPIEXEC = 'mpiexec'
DEFAULT_NUM_PROCESSES = 240

_fit_stage_1_log = logging.getLogger('allensdk.model.biophysical.fit_stage_1')

def find_core1_trace(data_set, all_sweeps):
    sweep_type = "C1LSCOARSE"
    _, sweeps, statuses = ephys_utils.get_sweeps_of_type(sweep_type, all_sweeps)
    sweep_status = dict(zip(sweeps, statuses))

    sweep_info = {}
    for s in sweeps:
        if sweep_status[s][-6:] == "failed":
            continue
        v, i, t = ephys_utils.get_sweep_v_i_t_from_set(data_set, s)
        if np.all(v[-100:] == 0): # Check for early termination of sweep
            continue
        stim_start, stim_dur, stim_amp, start_idx, end_idx = ephys_utils.get_step_stim_characteristics(i, t)
        swp = EphysSweepFeatureExtractor(t, v, i, start=stim_start, end=(stim_start + stim_dur))
        swp.process_spikes()
        isi_cv = swp.sweep_feature("isi_cv", allow_missing=True)
        sweep_info[s] = {"amp": stim_amp,
                         "n_spikes": len(swp.spikes()),
                         "quality": is_trace_good_quality(v, i, t),
                         "isi_cv": isi_cv}
    
    rheobase_amp = 1e12
    for s in sweep_info:
        if sweep_info[s]["amp"] < rheobase_amp and sweep_info[s]["n_spikes"] > 0:
            rheobase_amp = sweep_info[s]["amp"]
    sweep_to_use_amp = 1e12
    sweep_to_use_isi_cv = 1e121    
    sweep_to_use = -1
    for s in sweep_info:
        if sweep_info[s]["quality"] and sweep_info[s]["amp"] >= 39.0 + rheobase_amp and sweep_info[s]["isi_cv"] < 1.2 * sweep_to_use_isi_cv:
            use_new_sweep = False
            if sweep_to_use_isi_cv > 0.3 and ((sweep_to_use_isi_cv - sweep_info[s]["isi_cv"]) / sweep_to_use_isi_cv) >= 0.2:
                use_new_sweep = True
            elif sweep_info[s]["amp"] < sweep_to_use_amp:
                use_new_sweep = True
            if use_new_sweep:
                _fit_stage_1_log.info("now using sweep" + str(s))
                sweep_to_use = s
                sweep_to_use_amp = sweep_info[s]["amp"]
                sweep_to_use_isi_cv = sweep_info[s]["isi_cv"]
            
    if sweep_to_use == -1:
        _fit_stage_1_log.warn("Could not find appropriate core 1 sweep!")
        return []
    else:
        return [sweep_to_use]

def find_core2_trace(data_set, all_sweeps):
    sweep_type = "C2SQRHELNG"
    _, sweeps, statuses = ephys_utils.get_sweeps_of_type(sweep_type, all_sweeps)
    sweep_status = dict(zip(sweeps, statuses))
    amp_list = []
    core2_amps = {}
    for s in sweeps:
        if sweep_status[s][-6:] == "failed":
            continue
        v, i, t = ephys_utils.get_sweep_v_i_t_from_set(data_set, s)
        if np.all(v[-100:] == 0):
            continue
        stim_start, stim_dur, stim_amp, start_idx, end_idx = ephys_utils.get_step_stim_characteristics(i, t)
        if stim_start is np.nan:
            sweep_status[s] = "manual_failed"
            continue
        core2_amps[s] = stim_amp
        amp_list.append(stim_amp)
    core2_amp_counter = Counter(amp_list)
    common_amps = core2_amp_counter.most_common(3)
    best_amp = 0
    best_n_good = -1
    for amp, _ in common_amps:
        n_good = 0
        for k in core2_amps:
            if core2_amps[k] == amp and sweep_status[k][-6:] == "passed":
                v, i, t = ephys_utils.get_sweep_v_i_t_from_set(data_set, k)
                if is_trace_good_quality(v, i, t):
                    n_good += 1
        if n_good > best_n_good:
            best_n_good = n_good
            best_amp = amp
        elif n_good == best_n_good and amp < best_amp:
            best_amp = amp
    if best_n_good <= 1:
        return []
    else:
        sweeps_to_fit = []
        for k in core2_amps:
            if core2_amps[k] == best_amp and sweep_status[k][-6:] == "passed":
                sweeps_to_fit.append(k)
        return sweeps_to_fit

def is_trace_good_quality(v, i, t):
    stim_start, stim_dur, stim_amp, start_idx, end_idx = ephys_utils.get_step_stim_characteristics(i, t)
    swp = EphysSweepFeatureExtractor(t, v, i, start=stim_start, end=(stim_start + stim_dur))
    swp.process_spikes()

    spikes = swp.spikes()
    rate = swp.sweep_feature("avg_rate")

    if rate < 5.0:
        return False

    time_to_end = stim_start + stim_dur - spikes[-1]["threshold_t"]
    avg_end_isi = (((spikes[-1]["threshold_t"] - spikes[-2]["threshold_t"]) +
                   (spikes[-2]["threshold_t"] - spikes[-3]["threshold_t"])) / 2.0)

    if time_to_end > 2 * avg_end_isi:
        return False

    isis = np.diff([spk["threshold_t"] for spk in spikes])
    if check_for_pause(isis):
        return False

    return True


def check_for_pause(isis):
    if len(isis) <= 2:
        return False
    
    for i, isi in enumerate(isis[1:-1]):
        if isi > 3 * isis[i + 1 - 1] and isi > 3 * isis[i + 1 + 1]:
            return True
    return False


def collect_target_features(ft):
    min_std_dict = {
        'avg_rate': 0.5,
        'adapt': 0.001,
        'peak_v': 2.0,
        'trough_v': 2.0,
        'fast_trough_v': 2.0,
        'slow_trough_delta_v': 2.0,
        'slow_trough_delta_time': 0.05,
        'latency': 5.0,
        'isi_cv': 0.1,
        'mean_isi': 0.5,
        'first_isi': 1.0,
        'time_to_end': 50.0,
        'v_baseline': 2.0,
        'width': 0.0001,
        'upstroke': 50.0,
        'downstroke': 50.0,
        'upstroke_v': 2.0,
        'downstroke_v': 2.0,
        'threshold_v': 2.0,
        'peak_to_fast_tr_time': 0.0005,
        'phase_slope': 5.0,
    }

    target_features = []
    for k in ft:
        t = {"name": k, "mean": ft[k]["mean"], "stdev": ft[k]["stdev"]}
        if k in min_std_dict and min_std_dict[k] > ft[k]["stdev"]:
            t["stdev"] = min_std_dict[k]
        target_features.append(t)
    return target_features


def prepare_stage_1(description, passive_fit_data):
    output_directory = description.manifest.get_path('WORKDIR')
    neuronal_model_data = ju.read(description.manifest.get_path('neuronal_model_data'))
    specimen_data = neuronal_model_data['specimen']
    specimen_id = neuronal_model_data['specimen_id']
    is_spiny = not any(t['name'] == u'dendrite type - aspiny' for t in specimen_data['specimen_tags'])
    all_sweeps = specimen_data['ephys_sweeps']
    data_set = NwbDataSet(description.manifest.get_path('stimulus_path'))
    swc_path = description.manifest.get_path('MORPHOLOGY')
    
    if not os.path.exists(output_directory):
        os.makedirs(output_directory)

    ra = passive_fit_data['ra']
    cm1 = passive_fit_data['cm1']
    cm2 = passive_fit_data['cm2']

    # Check for fi curve shift to decide to use core1 or core2
    fi_shift, n_core2 = check_fi_shift.estimate_fi_shift(data_set, all_sweeps)
    fi_shift_threshold = 30.0
    sweeps_to_fit = []
    if abs(fi_shift) > fi_shift_threshold:
        _fit_stage_1_log.info("FI curve shifted; using Core 1")
        sweeps_to_fit = find_core1_trace(data_set, all_sweeps)
    else:
        sweeps_to_fit = find_core2_trace(data_set, all_sweeps)

        if sweeps_to_fit == []:
            _fit_stage_1_log.info("Not enough good Core 2 traces; using Core 1")
            sweeps_to_fit = find_core1_trace(data_set, all_sweeps)

    _fit_stage_1_log.debug("will use sweeps: " + str(sweeps_to_fit))

    jxn = -14.0

    t_set = []
    v_set = []
    i_set = []
    for s in sweeps_to_fit:
        v, i, t = ephys_utils.get_sweep_v_i_t_from_set(data_set, s)
        v += jxn
        stim_start, stim_dur, stim_amp, start_idx, end_idx = ephys_utils.get_step_stim_characteristics(i, t)
        t_set.append(t)
        v_set.append(v)
        i_set.append(i)
    ext = EphysSweepSetFeatureExtractor(t_set, v_set, i_set, start=stim_start, end=(stim_start + stim_dur))
    ext.process_spikes()

    ft = {}
    blacklist = ["isi_type"]
    for k in ext.sweeps()[0].spike_feature_keys():
        if k in blacklist:
            continue
        pair = {}
        pair["mean"] = float(ext.spike_feature_averages(k).mean())
        pair["stdev"] = float(ext.spike_feature_averages(k).std())
        ft[k] = pair

    # "Delta" features
    sweep_avg_slow_trough_delta_time = []
    sweep_avg_slow_trough_delta_v = []
    sweep_avg_peak_trough_delta_time = []
    for swp in ext.sweeps():
        threshold_t = swp.spike_feature("threshold_t")
        fast_trough_t = swp.spike_feature("fast_trough_t")
        slow_trough_t = swp.spike_feature("slow_trough_t")

        delta_t = slow_trough_t - fast_trough_t
        delta_t[np.isnan(delta_t)] = 0.
        sweep_avg_slow_trough_delta_time.append(np.mean(delta_t[:-1] / np.diff(threshold_t)))

        fast_trough_v = swp.spike_feature("fast_trough_v")
        slow_trough_v = swp.spike_feature("slow_trough_v")
        delta_v = fast_trough_v - slow_trough_v
        delta_v[np.isnan(delta_v)] = 0.
        sweep_avg_slow_trough_delta_v.append(delta_v.mean())

    ft["slow_trough_delta_time"] = {"mean": float(np.mean(sweep_avg_slow_trough_delta_time)),
                                    "stdev": float(np.std(sweep_avg_slow_trough_delta_time))}
    ft["slow_trough_delta_v"] = {"mean": float(np.mean(sweep_avg_slow_trough_delta_v)),
                                 "stdev": float(np.std(sweep_avg_slow_trough_delta_v))}

    baseline_v = float(ext.sweep_features("v_baseline").mean())
    passive_fit_data["e_pas"] = baseline_v
    for k in ext.sweeps()[0].sweep_feature_keys():
        pair = {}
        pair["mean"] = float(ext.sweep_features(k).mean())
        pair["stdev"] = float(ext.sweep_features(k).std())
        ft[k] = pair

    # Determine highest step to check for depolarization block
    noise_1_sweeps, _, _ = ephys_utils.get_sweeps_of_type("C1NSSEED_1", all_sweeps)
    noise_2_sweeps, _, _ = ephys_utils.get_sweeps_of_type("C1NSSEED_2", all_sweeps)
    step_sweeps, _, _ = ephys_utils.get_sweeps_of_type("C1LSCOARSE", all_sweeps)
    all_sweeps = noise_1_sweeps + noise_2_sweeps + step_sweeps
    max_i = 0
    for s in all_sweeps:
        try:
            v, i, t = ephys_utils.get_sweep_v_i_t_from_set(data_set, s['sweep_number'])
        except:
            pass
        if np.max(i) > max_i:
            max_i = np.max(i)
    max_i += 10 # add 10 pA
    max_i *= 1e-3 # convert to nA

    # ----------- Generate output and submit jobs ---------------

    # Set up directories
    # Decide which fit(s) we are doing
    if (is_spiny and ft["width"]["mean"] < 0.8) or (not is_spiny and ft["width"]["mean"] > 0.8):
        fit_types = ["f6", "f12"]
    elif is_spiny:
        fit_types = ["f6"]
    else:
        fit_types = ["f12"]

    for fit_type in fit_types:
        fit_type_dir = os.path.join(output_directory, fit_type)
        if not os.path.exists(fit_type_dir):
            os.makedirs(fit_type_dir)
        for seed in SEEDS:
            seed_dir = "{:s}/s{:d}".format(fit_type_dir, seed)
            if not os.path.exists(seed_dir):
                os.makedirs(seed_dir)

    # Collect and save data for target.json file
    target_dict = {}
    target_dict["passive"] = [{
        "ra": ra,
        "cm": { "soma": cm1, "axon": cm1, "dend": cm2 },
        "e_pas": baseline_v
    }]

    swc_data = pd.read_table(swc_path, sep='\s', comment='#', header=None)
    has_apic = False
    if APICAL_DENDRITE_TYPE in pd.unique(swc_data[1]):
        has_apic = True
        _fit_stage_1_log.info("Has apical dendrite")
    else:
        _fit_stage_1_log.info("Does not have apical dendrite")

    if has_apic:
        target_dict["passive"][0]["cm"]["apic"] = cm2

    target_dict["fitting"] = [{
        "junction_potential": jxn,
        "sweeps": sweeps_to_fit,
        "passive_fit_info": passive_fit_data,
        "max_stim_test_na": max_i,        
    }]

    target_dict["stimulus"] = [{
        "amplitude": 1e-3 * stim_amp,
        "delay": 1000.0,
        "duration": 1e3 * stim_dur
    }]

    target_dict["manifest"] = []
    target_dict["manifest"].append({"type": "file", "spec": swc_path, "key": "MORPHOLOGY"})

    target_dict["target_features"] = collect_target_features(ft)

    target_file = os.path.join(output_directory, 'target.json')
    ju.write(target_file, target_dict)

    # Create config.json for each fit type
    config_base_data = ju.read(os.path.join(FIT_BASE_DIR,
                                            'config_base.json'))


    jobs = []
    for fit_type in fit_types:
        config = config_base_data.copy()
        fit_type_dir = os.path.join(output_directory, fit_type)
        config_path = os.path.join(fit_type_dir, "config.json")

        config["biophys"][0]["model_file"] = [ target_file, config_path]
        if has_apic:
            fit_style_file = os.path.join(FIT_BASE_DIR, 'fit_styles', '%s_fit_style.json' % (fit_type))
        else:
            fit_style_file = os.path.join(FIT_BASE_DIR, "fit_styles", "%s_noapic_fit_style.json" % (fit_type))

        config["biophys"][0]["model_file"].append(fit_style_file)
        config["manifest"].append({"type": "dir", "spec": fit_type_dir, "key": "FITDIR"})
        ju.write(config_path, config)

        for seed in SEEDS:
            logfile = os.path.join(output_directory, fit_type, 's%d' % seed, 'stage_1.log')
            jobs.append({
                    'config_path': os.path.abspath(config_path),
                    'fit_type': fit_type,
                    'log': os.path.abspath(logfile),
                    'seed': seed,
                    'num_processes': DEFAULT_NUM_PROCESSES
                    })
    return jobs


def run_stage_1(jobs):
    for job in jobs:
        args = [MPIEXEC,
                '-np',
                str(job['num_processes']),
                sys.executable,
                '-m',
                optimize.__name__,
                str(job['seed']),
                job['config_path'],
                str(optimize.DEFAULT_NGEN),
                str(optimize.DEFAULT_MU)]
        _fit_stage_1_log.debug(args)
        with open(job['log'], "w") as outfile:
            subprocess.call(args, stdout=outfile)


from allensdk.api.queries.brain_observatory_api import BrainObservatoryApi
from allensdk.api.cache import cacheable
from allensdk.core.brain_observatory_cache import BrainObservatoryCache
import allensdk.internal.core.lims_utilities as lu
import os
import collections
import pandas as pd
import sys

sql_query_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'pre_release_sql')

with open(os.path.join(sql_query_dir, 'experiment_pre_release_query.sql'), 'r') as f:
    experiment_pre_release_query = f.read()

with open(os.path.join(sql_query_dir, 'container_pre_release_query.sql'), 'r') as f:
    container_pre_release_query = f.read()

with open(os.path.join(sql_query_dir, 'cell_specimens_pre_release_query.sql'), 'r') as f:
    cell_specimens_pre_release_query = f.read()

class BrainObservatoryApiPreRelease(BrainObservatoryApi):

    @cacheable()
    def get_experiment_containers(self):

        query_result = lu.query(container_pre_release_query)
        container_list = []
        for q in query_result:
            
            # # For development: print key/val pairs generated from LIMS query:
            # for key, val in sorted(q.items(), key=lambda x: x[0]):
            #     print key, val
            # raise

            c = collections.defaultdict(collections.defaultdict)

            c['id'] = q['ec_id']
            c['targeted_structure']['acronym'] = q['acronym']
            c['specimen']['donor'] = collections.defaultdict(collections.defaultdict)
            c['specimen']['donor']['external_donor_name'] = q['external_donor_name']
            c['specimen']['donor']['transgenic_lines'] = [collections.defaultdict(collections.defaultdict), collections.defaultdict(collections.defaultdict)]
            c['specimen']['donor']['transgenic_lines'][0]['transgenic_line_type_name'] = 'driver'
            c['specimen']['donor']['transgenic_lines'][0]['name'] = q['driver']
            c['specimen']['donor']['transgenic_lines'][1]['transgenic_line_type_name'] = 'reporter'
            c['specimen']['donor']['transgenic_lines'][1]['name'] = q['reporter']
            c['specimen']['name'] = q['specimen']
            c['imaging_depth'] = q['depth']
            c['failed'] = q['oa_state'] == 'failed'


            if q['donor_tags'] == u'Epileptiform Events':
                c['specimen']['donor']['conditions'] = [collections.defaultdict(collections.defaultdict)]
                c['specimen']['donor']['conditions'][0]['name'] = u'Epileptiform Events'
            elif q['donor_tags'] == u'':
                pass
            else:
                raise

            container_list.append(c)
        return container_list


    @cacheable()
    def get_ophys_experiments(self):

        query_result = lu.query(experiment_pre_release_query)
        experiment_list = []
        for q in query_result:
            c = collections.defaultdict(collections.defaultdict)

            # # For development: print key/val pairs generated from LIMS query:
            # for key, val in sorted(q.items(), key=lambda x: x[0]):
            #     print key, val
            # raise

            c['id'] = q['o_id']
            c['imaging_depth'] = q['depth']
            c['targeted_structure']['acronym'] = q['acronym']
            c['specimen']['donor'] = collections.defaultdict(collections.defaultdict)
            c['specimen']['donor']['external_donor_name'] = q['acronym']
            c['specimen']['donor']['transgenic_lines'] = [collections.defaultdict(collections.defaultdict), collections.defaultdict(collections.defaultdict)]
            c['specimen']['donor']['transgenic_lines'][0]['transgenic_line_type_name'] = 'driver'
            c['specimen']['donor']['transgenic_lines'][0]['name'] = q['driver']
            c['specimen']['donor']['transgenic_lines'][1]['transgenic_line_type_name'] = 'reporter'
            c['specimen']['donor']['transgenic_lines'][1]['name'] = q['reporter']
            c['date_of_acquisition'] = q['date_of_acquisition']
            c['specimen']['donor']['date_of_birth'] = q['date_of_birth']
            c['experiment_container_id'] = q['ec_id']
            c['stimulus_name'] = q['stimulus_name']
            c['specimen']['donor']['external_donor_name'] = q['external_donor_name']
            c['specimen']['name'] = q['specimen']
            c['fail_eye_tracking'] = q['fail_eye_tracking']

            experiment_list.append(c)
        return experiment_list

    @cacheable()
    def get_cell_metrics(self):
        query_result = lu.query(cell_specimens_pre_release_query)

        mappings = self.get_stimulus_mappings()
        thumbnails = [m['item'] for m in mappings if m['item_type'] == 'T' and m['level'] == 'R']

        cell_list = []
        for q in query_result:
            c = collections.defaultdict(collections.defaultdict)

            c['all_stim'] = q['a_valid'] and q['b_valid'] and q['c_valid']
            for key in ['cell_specimen_id', 'area', 'donor_full_genotype', 'experiment_container_id', 'imaging_depth', 'specimen_id', 'tld1_id',
                        'tld1_name', 'tld2_id', 'tld2_name', 'tlr1_id', 'tlr1_name']:
                c[key] = q[key]

            if q['failed_experiment_container'] == 't':
                c['failed_experiment_container'] = True
            elif q['failed_experiment_container'] == 'f':
                c['failed_experiment_container'] = False
            else:
                raise RuntimeError('Unexpected value: {} not in ("t", "f")'.format(q['failed_experiment_container']))
            

            # Session A metrics:
            for key in ['dsi_dg', 'g_dsi_dg', 'g_osi_dg', 'osi_dg', 'p_dg', 'p_run_mod_dg', 'peak_dff_dg', 'pref_dir_dg', 'pref_tf_dg', 'reliability_dg',
                        'reliability_nm3', 'run_mod_dg', 'tfdi_dg', 'tfdi_dg']:
                if q['crara_data'] is None:
                    c[key] = None
                else:
                    c[key] = q['crara_data']['roi_cell_metrics'].get(key,None)

            # Session B metrics:
            for key in ['g_osi_sg', 'image_sel_ns', 'osi_sg', 'p_ns', 'p_run_mod_ns', 'p_run_mod_sg', 'p_sg', 'peak_dff_ns', 'peak_dff_sg', 'pref_image_ns',
                        'pref_ori_sg', 'pref_phase_sg', 'pref_sf_sg', 'pref_image_ns', 'pref_ori_sg', 'reliability_ns', 'reliability_sg',
                        'run_mod_ns','run_mod_sg','sfdi_sg', 'time_to_peak_ns', 'time_to_peak_sg']:
                
                if q['crarb_data'] is None:
                    c[key] = None
                else:
                    c[key] = q['crarb_data']['roi_cell_metrics'].get(key,None)

            # Session C metrics:
            for key in ['reliability_nm2', 'rf_area_off_lsn', 'rf_area_on_lsn', 'rf_center_off_x_lsn', 'rf_center_off_y_lsn',
                        'rf_center_on_x_lsn', 'rf_center_on_y_lsn', 'rf_chi2_lsn', 'rf_distance_lsn', 'rf_overlap_index_lsn',
                        ]:
                if q['crarc_data'] is None:
                    c[key] = None
                else:
                    c[key] = q['crarc_data']['roi_cell_metrics'].get(key,None)
            
            for suffix in ['a', 'b', 'c']:
                if not q['crar%s_data' % suffix] is None:
                    c['reliability_nm1_%s' % suffix] = q['crar%s_data' % suffix]['roi_cell_metrics'].get('reliability_nm1',None)
                else:
                    c['reliability_nm1_%s' % suffix] = None
            
            # Fake in thumbnail images:
            for t in thumbnails:
                c[t] = None

            # # For development: print key/val pairs generated from LIMS query:
            # for key, val in sorted(q.items(), key=lambda x: x[0]):
            #     if key == 'crarb_data':
            #         print 'crarb_data[roi_cell_metrics]'
            #         for key2, val2 in sorted(val['roi_cell_metrics'].items(), key=lambda x: x[0]):
            #             print '    ', key2, val2
            #     else:
            #         print key, val
            # raise

            cell_list.append(c)

        return cell_list


import matplotlib
matplotlib.use('agg')
import matplotlib.pyplot as plt

import allensdk.internal.core.lims_utilities as lu
from allensdk.internal.core.lims_pipeline_module import PipelineModule, run_module

import argparse, os, logging, shutil
import h5py
import numpy as np
import shutil

import allensdk.internal.brain_observatory.demixer as demixer
from allensdk.config.manifest import Manifest

import allensdk.core.json_utilities as ju
import logging

EXCLUDE_LABELS = ["union", "duplicate", "motion_border" ]

def debug(experiment_id, local=False):
    OUTPUT_DIRECTORY = "/data/informatics/CAM/demix"
    SDK_PATH = "/data/informatics/CAM/analysis/allensdk"
    SCRIPT = "/data/informatics/CAM/analysis/allensdk/allensdk/internal/pipeline_modules/run_demixing.py"

    sd = lu.query("select storage_directory from ophys_experiments where id = %d" % experiment_id)[0]['storage_directory']
    rois = lu.query("select * from cell_rois where ophys_experiment_id = %d" % experiment_id)

    exc_labels = lu.query("""
select cr.id, rel.name as exclusion_label from cell_rois cr
join cell_rois_roi_exclusion_labels crrel on crrel.cell_roi_id = cr.id
join roi_exclusion_labels rel on crrel.roi_exclusion_label_id = rel.id
where cr.ophys_experiment_id = %d
""" % experiment_id)

    nrois = { roi['id']: dict(width=roi['width'],
                              height=roi['height'], 
                              x=roi['x'],
                              y=roi['y'],
                              id=roi['id'],
                              valid=roi['valid_roi'],
                              mask=roi['mask_matrix'],
                              exclusion_labels=[]) 
              for roi in rois }

    for exc_label in exc_labels:
        nrois[exc_label['id']]['exclusion_labels'].append(exc_label['exclusion_label'])
        
    exp_dir = os.path.join(OUTPUT_DIRECTORY, str(experiment_id))

    input_data = {
        "movie_h5": os.path.join(sd, "processed", "concat_31Hz_0.h5"),
        "traces_h5": os.path.join(sd, "processed", "roi_traces.h5"),
        "roi_masks": nrois.values(),
        "output_file": os.path.join(exp_dir, "demixed_traces.h5")
        }
 
    run_module(SCRIPT, 
               input_data, 
               exp_dir,
               sdk_path=SDK_PATH,
               pbs=dict(vmem=160,
                        job_name="demix_%d"% experiment_id,
                        walltime="36:00:00"),
               local=local,
               optional_args=['--log-level','DEBUG'])

def assert_exists(file_name):
    if not os.path.exists(file_name):
        raise IOError("file does not exist: %s" % file_name)


def get_path(obj, key, check_exists):
    try:
         path = obj[key]
    except KeyError:
        raise KeyError("required input field '%s' does not exist" % key)

    if check_exists:
        assert_exists(path)

    return path


def parse_input(data, exclude_labels):
    movie_h5 = get_path(data, "movie_h5", True)
    traces_h5 = get_path(data, "traces_h5", True)
    output_h5 = get_path(data, "output_file", False)

    with h5py.File(movie_h5, "r") as f:
        movie_shape = f["data"].shape[1:]

    with h5py.File(traces_h5, "r") as f:
        traces = f["data"].value
        trace_ids = [ int(rid) for rid in f["roi_names"].value ]

    rois = get_path(data, "roi_masks", False)
    masks = None
    valid = None

    for roi in rois:
        mask = np.zeros(movie_shape, dtype=bool)
        mask_matrix = np.array(roi["mask"], dtype=bool)
        mask[roi["y"]:roi["y"]+roi["height"],roi["x"]:roi["x"]+roi["width"]] = mask_matrix
        
        if masks is None:
            masks = np.zeros((len(rois), mask.shape[0], mask.shape[1]), dtype=bool)
            valid = np.zeros(len(rois), dtype=bool)

        rid = int(roi["id"])
        try:
            ridx = trace_ids.index(rid)
        except ValueError as e:
            raise ValueError("Could not find cell roi id %d in roi traces file" % rid)

        masks[ridx,:,:] = mask

        valid[ridx] = len(set(exclude_labels) & set(roi.get("exclusion_labels",[]))) == 0

    return traces, masks, valid, np.array(trace_ids), movie_h5, output_h5

def main():
    mod = PipelineModule()
    mod.parser.add_argument("--exclude-labels", nargs="*", default=EXCLUDE_LABELS)

    data = mod.input_data()
    logging.debug("reading input")

    traces, masks, valid, trace_ids, movie_h5, output_h5 = parse_input(data, mod.args.exclude_labels)
    
    logging.debug("excluded masks: %s", str(zip(np.where(~valid)[0], trace_ids[~valid])))
    output_dir = os.path.dirname(output_h5)
    plot_dir = os.path.join(output_dir, "demix_plots")
    if os.path.exists(plot_dir):
        shutil.rmtree(plot_dir)
    Manifest.safe_mkdir(plot_dir)

    logging.debug("reading movie")
    with h5py.File(movie_h5, 'r') as f:
        movie = f['data'].value

    # only demix non-union, non-duplicate ROIs
    valid_idxs = np.where(valid)
    demix_traces = traces[valid_idxs]
    demix_masks = masks[valid_idxs]

    logging.debug("demixing")
    demixed_traces, drop_frames = demixer.demix_time_dep_masks(demix_traces, movie, demix_masks)
    
    nt_inds = demixer.plot_negative_transients(demix_traces, 
                                               demixed_traces, 
                                               valid[valid_idxs],
                                               demix_masks,
                                               trace_ids[valid_idxs],
                                               plot_dir)

    logging.debug("rois with negative transients: %s", str(trace_ids[valid_idxs][nt_inds]))

    nb_inds = demixer.plot_negative_baselines(demix_traces, 
                                              demixed_traces, 
                                              demix_masks,
                                              trace_ids[valid_idxs],
                                              plot_dir)

    # negative baseline rois (and those that overlap with them) become nans 
    logging.debug("rois with negative baselines (or overlap with them): %s", str(trace_ids[valid_idxs][nb_inds]))
    demixed_traces[nb_inds, :] = np.nan

    logging.info("Saving output")    
    out_traces = np.zeros(traces.shape, dtype=demix_traces.dtype)
    out_traces[:] = np.nan
    out_traces[valid_idxs] = demixed_traces

    with h5py.File(output_h5, 'w') as f:
        f.create_dataset("data", data=out_traces, compression="gzip")
        f.create_dataset("roi_names", data=[str(rn) for rn in trace_ids ])

    mod.write_output_data(dict(
            negative_transient_roi_ids=trace_ids[valid_idxs][nt_inds],
            negative_baseline_roi_ids=trace_ids[valid_idxs][nb_inds]
            ))


if __name__ == "__main__": main()



import itertools
import numpy as np
import logging

class MaskSet( object ):
    def __init__(self, masks):
        self.masks = masks
        self.bbs = make_bbs(self.masks)
        self.mask_dist = bb_dist(self.bbs)

        self.cached_sizes = {}

        self.cached_unions = {}
        self.cached_union_sizes = {}

        self.cached_intersections = {}
        self.cached_intersection_sizes = {}

    @property
    def count(self):
        return len(self.bbs)

    def distance(self, mask_idxs):
        return max(self.mask_dist[i,j] for (i,j) in itertools.combinations(mask_idxs, 2))

    def close(self, mask_idxs, max_dist):
        return not any(self.mask_dist[i,j] > max_dist for (i,j) in itertools.combinations(mask_idxs, 2)) 

    def close_sets(self, set_size, max_dist):
        mask_sets = itertools.combinations(range(len(self.bbs)), set_size)
        return (ms for ms in mask_sets if self.close(ms,  max_dist))

    def _idx_key(self, idxs):
        return tuple(sorted(set(idxs)))

    def mask(self, mask_idx):
        return self.masks[mask_idx]

    def union(self, mask_idxs):
        mask_idxs = self._idx_key(mask_idxs)
        
        if mask_idxs in self.cached_unions:
            return self.cached_unions[mask_idxs]

        if len(mask_idxs) == 0:
            return None

        i0 = mask_idxs[0]
        union = self.masks[i0].copy()
        
        if len(mask_idxs) == 1:
            return union

        for idx in mask_idxs[1:]:
            union |= self.masks[idx]

        self.cached_unions[mask_idxs] = union

        return union

    def overlap_fraction(self, idx0, idx1):
        union_size = self.union_size([idx0,idx1])
        overlap_size = self.intersection_size([idx0,idx1])
        return float(overlap_size) / float(union_size)

    def detect_duplicates(self, overlap_threshold):
        duplicate_masks = set()

        for idx0,idx1 in self.close_sets(set_size=2, max_dist=0):
            overlap_frac = self.overlap_fraction(idx0, idx1)

            if overlap_frac > overlap_threshold:
                duplicate_masks.add(tuple(sorted([idx0,idx1])))

        return duplicate_masks

    def mask_is_union_of_set(self, mask_idx, set_idxs, threshold):
        # does this mask overlap with each element of the set individually?
        # i.e. overlap of mask and set element covers most of the set element
        for set_mask_idx in set_idxs:
            overlap_size = self.intersection_size([set_mask_idx, mask_idx])
            set_mask_size = self.size(set_mask_idx)
            if overlap_size < threshold * set_mask_size:
                return False


        # does this mask cover more than the union of the individual set elements?
        set_union = self.union(set_idxs)
        mask = self.mask(mask_idx)
        overlap = set_union & mask
        overlap_size = overlap.sum()

        return overlap_size > self.size(mask_idx) * threshold

    def detect_unions(self, set_size=2, max_dist=10, threshold=0.7):
        union_masks = {}

        mask_combos = list(self.close_sets(set_size, max_dist))

        for i, set_idxs in enumerate(mask_combos):
            for mask_idx in range(self.count):
                if mask_idx in set_idxs:
                    continue
                elif not self.close([mask_idx] + list(set_idxs), max_dist):
                    continue        

                if self.mask_is_union_of_set(mask_idx, set_idxs, threshold):                    
                    if mask_idx in union_masks:
                        logging.warning("already detected this mask as a union")
                    union_masks[mask_idx] = set_idxs

        return union_masks

    def union_size(self, mask_idxs):
        mask_idxs = self._idx_key(mask_idxs)

        if mask_idxs in self.cached_union_sizes:
            return self.cached_union_sizes[mask_idxs]
   
        s = self.union(mask_idxs).sum()
        self.cached_union_sizes[mask_idxs] = s

        return s

    def intersection(self, mask_idxs):        
        mask_idxs = self._idx_key(mask_idxs)
        
        if mask_idxs in self.cached_intersections:
            return self.cached_intersections[mask_idxs]

        if len(mask_idxs) == 0:
            return None

        # don't cache the empty ones
        if not self.close(mask_idxs, 0):
            return np.zeros(self.masks[0].shape)
        
        i0 = mask_idxs[0]
        intersection = self.masks[i0].copy()

        if len(mask_idxs) == 1:
            return intersection

        for idx in mask_idxs[1:]:
            intersection &= self.masks[idx]

        self.cached_intersections[mask_idxs] = intersection

        return intersection

    def intersection_size(self, mask_idxs):
        mask_idxs = self._idx_key(mask_idxs)

        if mask_idxs in self.cached_intersection_sizes:
            return self.cached_intersection_sizes[mask_idxs]
   
        s = self.intersection(mask_idxs).sum()
        self.cached_intersection_sizes[mask_idxs] = s

        return s

    def size(self, mask_idx):
        return self.union_size([mask_idx])


def make_bbs(masks):
    bbs = []

    for i in range(len(masks)):
        m = np.where(masks[i])
        bbs.append([[m[0].min(), m[0].max()],[m[1].min(), m[1].max()]])

    return bbs

def bb_dist(bbs):
    num_bbs = len(bbs)

    dist = np.zeros((num_bbs, num_bbs))
    for i,j in itertools.combinations(range(num_bbs), 2):
        bbi = bbs[i]
        bbj = bbs[j]

        if bbi[0][0] < bbj[0][1]:
            distx = bbj[0][0] - bbi[0][1]
        else:
            distx = bbi[0][0] - bbj[0][1]

        if bbi[1][0] < bbj[1][1]:
            disty = bbj[1][0] - bbi[1][1]
        else:
            disty = bbi[1][0] - bbj[1][1]

        dist[i,j] = max(distx,disty)
        dist[j,i] = dist[i,j]

    return dist



# Copyright 2016 Allen Institute for Brain Science
# This file is part of Allen SDK.
#
# Allen SDK is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# Allen SDK is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Allen SDK.  If not, see <http://www.gnu.org/licenses/>.

import json
import math

def euclidean_distance(node1, node2):
    dx = node1.x - node2.x
    dy = node1.y - node2.y
    dz = node1.z - node2.z
    return math.sqrt(dx*dx + dy*dy + dz*dz)


def midpoint(node1, node2):
    px = (node1.x + node2.x) * 0.5
    py = (node1.y + node2.y) * 0.5
    pz = (node1.z + node2.z) * 0.5
    return [px, py, pz]


class Node(object): 
    """
    Represents node in SWC morphology file
    """

    def __init__(self, n, t, x, y, z, r, pn, **kwargs):
        """
        Parameters
        ----------
        n: integer
        node ID

        t: integer
        node type (SOMA, AXON, BASAL_DENDRITE or APICAL_DENDRITE)

        x: float
        x position of node

        y: float
        y position of node

        z: float
        z position of node

        r: float
        radius of node

        pn: integer
        ID of parent node (-1 if no parent)
        """
        # these values correspond to columns in an SWC file
        self.n = n
        self.t = t
        self.x = x
        self.y = y
        self.z = z
        self.radius = r
        self.parent = pn
        # 
        self.children = []  # IDs of child nodes
        self.tree_id = -1      # which unconnected graph this node belongs to
        # number of compartment that has this node as its endpoint
        # all nodes except root nodes have a compartment
        self.compartment_id = -1    

    def to_dict(self):
        """ Convert the node into a serializable dictionary """
        return {
            "id": self.n,
            "type": self.t,
            "x": self.x,
            "y": self.y,
            "z": self.z,
            "radius": self.radius,
            "parent": self.parent,
            "children": self.children,
            "tree_id": self.tree_id,
            "compartment_id": self.compartment_id
        }

    @classmethod
    def from_dict(cls, d):
        return cls(
            n = d["id"],
            t = d["type"],
            x = d["x"],
            y = d["y"],
            z = d["z"],
            r = d["radius"],
            pn = d["parent"],
        )

    def __getitem__(self, item):
        return self.to_dict()[item]

    def __str__(self):
        return self.short_string()
        return json.dumps(self.to_dict())

    def short_string(self):
        """ create string with node information in succinct, 
        single-line form """
        return "%d %d %.4f %.4f %.4f %.4f %d %s %d" % (self.n, self.t, self.x, self.y, self.z, self.radius, self.parent,
        str(self.children), self.tree_id);

# Morphology nodes have the following fields. These allow dictionary access
#   to node fields (this is for backward compatibility)
NODE_ID      = 'id'
NODE_TYPE    = 'type'
NODE_X       = 'x'
NODE_Y       = 'y'
NODE_Z       = 'z'
NODE_R       = 'radius'
NODE_PN      = 'parent'
NODE_TREE_ID = 'tree_id'     
NODE_CHILDREN = 'children'   




import numpy as np
import logging

MONITOR_POSITION_OLD = np.array([17.0, 0.0, 0.0])
MONITOR_POSITION_NEW = np.array([11.86, 8.62, 3.16])
MONITOR_ROTATIONS = np.array([0.0, 0.0, 0.0])

CAMERA_POSITION_OLD = np.array([13.0, 0, 0])
CAMERA_POSITION_NEW = np.array([10.28, 7.47, 2.74])
CAMERA_ROTATIONS_OLD = np.array([0.0, 0.0, 13.1*np.pi/180])
CAMERA_ROTATIONS_NEW = np.array([0.0, 0.0, 2.8*np.pi/180])

LED_POSITION_ORIGINAL = np.array([26.51, -3.93, 0.1])
LED_POSITION_OLD = np.array([25.89, -6.12, 3.21])
LED_POSITION_NEW = np.array([24.6, 9.23, 5.26])

EYE_RADIUS = 0.1682  # in cm
CM_PER_PIXEL = 10.2/10000.0

class EyeCalibration(object):
    '''Class for performing eye-tracking calibration.

    Provides methods for estimating the position of the pupil in
    3D space and projecting the gaze onto the monitor in both
    3D space and monitor space given the experimental geometry.

    Parameters
    ----------
    monitor_position : numpy.ndarray
        [x,y,z] position of monitor in cm.
    monitor_rotations : numpy.ndarray
        [x,y,z] rotations of monitor in radians.
    led_position : numpy.ndarray
        [x,y,z] position of LED in cm.
    camera_position : numpy.ndarray
        [x,y,z] position of camera in cm.
    camera_rotations : numpy.ndarray
        [x,y,z] rotations for camera in radians. X and Y must be 0.
    eye_radius : float
        Radius of the eye in cm.
    cm_per_pixel : float
        Pixel size of eye-tracking camera.
    '''
    def __init__(self, monitor_position=MONITOR_POSITION_NEW,
                 monitor_rotations=MONITOR_ROTATIONS,
                 led_position=LED_POSITION_OLD,
                 camera_position=CAMERA_POSITION_OLD,
                 camera_rotations=CAMERA_ROTATIONS_OLD,
                 eye_radius=EYE_RADIUS,
                 cm_per_pixel=CM_PER_PIXEL):
        '''Constructor.'''

        self.eye_radius = eye_radius
        self.cm_per_pixel = cm_per_pixel

        self.monitor_position = monitor_position
        self.led_position = led_position
        self.camera_position = camera_position

        self.cr = self.cr_position_in_mouse_eye_coordinates(led_position,
                                                            eye_radius)

        self.monitor_rotations = monitor_rotations
        if camera_rotations[0] != 0 or camera_rotations[1] != 0:
            logging.warning("Got nonzero x=%s,y=%s rotations for camera",
                            camera_rotations[0], camera_rotations[1])
        self.camera_rotation = camera_rotations[2]

    def pupil_position_in_mouse_eye_coordinates(self, pupil_parameters,
                                                cr_parameters):
        '''Compute the 3D pupil position in mouse eye coordinates.

        Parameters
        ----------
        pupil_parameters :  numpy.ndarray
            Array of pupil parameters for each eye tracking frame.
        cr_paramaeters : numpy.ndarray
            Array of corneal reflection parameters for each eye
            tracking frame.

        Returns
        -------
        numpy.ndarray
            Pupil position estimates in eye coordinates.
        '''
        # x, y are in screen coordinates, with y increasing towards the top
        delta_px = (pupil_parameters.T[0] - cr_parameters.T[0]) * \
            self.cm_per_pixel
        delta_py = (cr_parameters.T[1] - pupil_parameters.T[1]) * \
            self.cm_per_pixel  # +y is down on image

        R_cam_to_eye = base_object_to_eye_rotation_matrix(
            self.camera_position)
        # camera frame is passed to us pointed at the eye, but the image
        # appears as if the camera were rotated 180 degrees about its y-axis
        R_cam = object_rotation_matrix(0, np.pi, self.camera_rotation)
        cr_cam = np.dot(R_cam, np.dot(R_cam_to_eye.T, self.cr))

        px_cam = cr_cam[0] + delta_px
        py_cam = cr_cam[1] + delta_py
        pz_cam = np.sqrt(self.eye_radius**2 - px_cam**2 - py_cam**2)

        # estimating a position outside the eyeball is impossible, bad data
        bad_idx = np.sqrt(px_cam**2 + py_cam**2) > self.eye_radius
        px_cam[bad_idx] = np.nan
        py_cam[bad_idx] = np.nan
        pz_cam[bad_idx] = np.nan

        p_cam = np.vstack([px_cam, py_cam, pz_cam])

        # rotate estimates 
        return np.dot(R_cam_to_eye, np.dot(R_cam.T, p_cam)).T

    @staticmethod
    def cr_position_in_mouse_eye_coordinates(led_position, eye_radius):
        '''Determine the 3D position of the corneal reflection.

        The eye is modeled as a spherical mirror, so the reflection
        appears to be half the radius of the eye from the origin along
        the eye-LED axis.

        Parameters
        ----------
        led_position : numpy.ndarray
            [x,y,z] position of the LED in eye coordinates.
        eye_radius : float
            Radius of the eye in centimeters.

        Returns
        -------
        numpy.ndarray
            [x,y,z] location of the corneal reflection in eye coordinates.
        '''
        return (eye_radius/(2*np.linalg.norm(led_position))) * led_position

    def pupil_position_on_monitor_in_cm(self, pupil_parameters,
                                        cr_parameters):
        '''Compute the pupil position on the monitor in cm.

        Parameters
        ----------
        pupil_parameters :  numpy.ndarray
            Array of pupil parameters for each eye tracking frame.
        cr_paramaeters : numpy.ndarray
            Array of corneal reflection parameters for each eye
            tracking frame.

        Returns
        -------
        numpy.ndarray
            Pupil position estimates in eye coordinates.
        '''
        pupil_positions = self.pupil_position_in_mouse_eye_coordinates(
            pupil_parameters, cr_parameters)

        monitor_normal = object_norm_eye_coordinates(
            self.monitor_position, self.monitor_rotations[0],
            self.monitor_rotations[1], self.monitor_rotations[2])

        projected_positions = project_to_plane(monitor_normal,
                                               self.monitor_position,
                                               pupil_positions)

        monitor_positions = projected_positions - self.monitor_position

        R_monitor_to_eye = base_object_to_eye_rotation_matrix(
            self.monitor_position)
        R_monitor = object_rotation_matrix(self.monitor_rotations[0],
                                           self.monitor_rotations[1],
                                           self.monitor_rotations[2])

        result = np.dot(R_monitor.T,
                        np.dot(R_monitor_to_eye.T, monitor_positions.T))
        return result[:2].T

    def pupil_position_on_monitor_in_degrees(self, pupil_parameters,
                                             cr_parameters):
        '''Get pupil position on monitor measured in visual degrees.

        Parameters
        ----------
        pupil_parameters :  numpy.ndarray
            Array of pupil parameters for each eye tracking frame.
        cr_paramaeters : numpy.ndarray
            Array of corneal reflection parameters for each eye
            tracking frame.

        Returns
        -------
        numpy.ndarray
            Pupil position estimate in visual degrees.
        '''

        mag = np.sqrt(np.sum(self.monitor_position**2))

        pupil_pos = self.pupil_position_on_monitor_in_cm(pupil_parameters,
                                                         cr_parameters)

        x = pupil_pos.T[0]
        y = pupil_pos.T[1]

        meridian = np.arctan(x/mag)*180/np.pi
        elevation = np.arctan(y/np.sqrt(mag**2 + x**2))*180/np.pi

        angles = np.vstack([meridian, elevation]).T

        return angles

    def compute_area(self, pupil_parameters):
        '''Compute the area of the pupil.

        Assume the pupil is a circle, and that as it moves off-axis
        with the camera the observed ellipse major axis remains the
        diameter of the circle.

        Parameters
        ----------
        pupil_parameters : numpy.ndarray
            [nx5] array of pupil parameters.

        Returns
        -------
        numpy.ndarray
            [nx1] array of pupil areas in estimated pixels.
        '''
        r = np.maximum(pupil_parameters.T[3], pupil_parameters.T[4])
        return np.pi*r*r


def project_to_plane(plane_normal, plane_point, points):
    '''Project from the origin through points onto a plane.

    Parameters
    ----------
    plane_normal : numpy.ndarray
        [x, y, z] normal unit vector to the plane.
    plane_point : numpy.ndarray
        [x, y, z] point on the plane.
    points : numpy.ndarray
        [nx3] points in space through which to project.

    Returns
    -------
    numpy.ndarray
        [nx3] points projected on the plane.
    '''
    factor = np.sum(plane_normal*plane_point) / \
             np.sum(plane_normal*points, axis=1)
    return (factor*points.T).T


def object_norm_eye_coordinates(object_position, x_rotation,
                                y_rotation, z_rotation):
    '''Get the normal vector for the object plane in eye coordinates.

    Parameters
    ----------
    object_position : numpy.ndarray
        [x, y, z] location of the object in eye coordinates.
    x_rotation : float
        Rotation about the x-axis in radians.
    y_rotation : float
        Rotation about the y-axis in radians.
    z_rotation : float
        Rotation about the z-axis in radians.

    Returns
    -------
    numpy.ndarray
        Endpoint of the object plane vector in eye coordinates.
    '''
    R_object_to_eye = base_object_to_eye_rotation_matrix(object_position)
    R_object_frame = object_rotation_matrix(x_rotation, y_rotation,
                                            z_rotation)
    return np.dot(R_object_to_eye, np.dot(R_object_frame, [0, 0, 1]))


def base_object_to_eye_rotation_matrix(object_position):
    '''Rotation matrix to rotate base object frame to eye coordinates.

    By convention, any other object's coordinate frame before rotations
    is set with positive Z pointing from the object's position back
    to the origin of the eye coordinate system, with X parallel to the
    eye X-Y plane.

    Parameters
    ----------
    object_position : np.ndarray
        [x, y, z] position of object in eye coordinates.

    Returns
    -------
    numpy.ndarray
        [3x3] rotation matrix.
    '''
    eye_norm = -object_position/np.linalg.norm(object_position)

    # rotate about eye-z to align eye-x to object-x
    theta_z = -(np.pi/2 + np.arctan2(eye_norm[1], eye_norm[0]))
    Rz = np.array([[np.cos(theta_z), -np.sin(theta_z), 0],
                   [np.sin(theta_z), np.cos(theta_z), 0],
                   [0, 0, 1]])
    eye_norm_about_z = np.dot(Rz, eye_norm)

    # rotate about x' to align eye-z to object-z
    theta_x = np.pi/2 - np.arctan2(eye_norm_about_z[2], eye_norm_about_z[1])
    Rx = np.array([[1, 0, 0],
                   [0, np.cos(theta_x), -np.sin(theta_x)],
                   [0, np.sin(theta_x), np.cos(theta_x)]])

    R = np.dot(Rx, Rz).T
    return R


def object_rotation_matrix(x_rotation, y_rotation, z_rotation):
    '''Rotation matrix in object coordinate frame.

    The rotation matrix for rotating the object coordinate frame from
    the initial position. This is done by rotating around x, then
    around y', then around z''.

    Parameters
    ----------
    x_rotation : float
        Rotation about x axis in radians.
    y_rotation : float
        Rotation about y axis in radians.
    z_rotation : float
        Rotation about z axis in radians.

    Returns
    -------
    numpy.ndarray
        [3x3] rotation matrix.
    '''
    Rx = np.array([[1, 0, 0],
                   [0, np.cos(x_rotation), -np.sin(x_rotation)],
                   [0, np.sin(x_rotation), np.cos(x_rotation)]])
    Ry = np.array([[np.cos(y_rotation), 0, np.sin(y_rotation)],
                   [0, 1, 0],
                   [-np.sin(y_rotation), 0, np.cos(y_rotation)]])
    Rz = np.array([[np.cos(z_rotation), -np.sin(z_rotation), 0],
                   [np.sin(z_rotation), np.cos(z_rotation), 0],
                   [0, 0, 1]])
    result = np.dot(Rz, np.dot(Ry, Rx))
    return result



import logging

from allensdk.internal.core.lims_pipeline_module import PipelineModule
from allensdk.internal.mouse_connectivity.interval_unionize.run_tissuecyte_unionize_cav import run 




def main():

    module = PipelineModule()

    input_data = module.input_data()
    output_data = run(input_data)
    module.write_output_data(output_data)


if __name__ == '__main__':
    main()



from __future__ import division

import logging

import SimpleITK as sitk
import numpy as np


def sitk_get_image_parameters(volume):
    return (np.array(volume.GetSpacing()), 
            np.array(volume.GetSize()), 
            np.array(volume.GetOrigin()))


def sitk_get_center(volume):
    _, size, _ = sitk_get_image_parameters(volume)
    return (size - 1) / 2


def sitk_get_size_parity(volume):
    _, size, _ = sitk_get_image_parameters(volume)
    return np.mod(size, 2) 


def sitk_get_diagonal_length(volume):
    _, size, _ = sitk_get_image_parameters(volume)
    return np.linalg.norm(size)


def sitk_paste_into_center(smaller, larger):
  
    smaller_parities = sitk_get_size_parity(smaller)
    larger_parities = sitk_get_size_parity(larger)
    if not np.allclose(smaller_parities, larger_parities):
        logging.warn('parities differ, result will not be centered : {0}, {1}'.format(smaller_parities, larger_parities))

    smaller_center = sitk_get_center(smaller)
    larger_center = sitk_get_center(larger)

    offset = np.around(larger_center - smaller_center).astype(int).tolist()
    return sitk.Paste(larger, smaller, smaller.GetSize(), [0, 0, 0,], offset)



import numpy as np
from allensdk.ephys.extract_cell_features import get_square_stim_characteristics
from scipy import stats
from scipy.optimize import curve_fit
import matplotlib.pyplot as plt

def MLIN(voltage, current, res, cap, dt, MAKE_PLOT=False, SHOW_PLOT=False, BLOCK=False, PUBLICATION_PLOT=False):
    '''voltage, current
    input:
        voltage: numpy array of voltage with test pulse cut out 
        current: numpy array of stimulus with test pulse cut out '''
    t = np.arange(0, len(current)) * dt
    (_, _, _, start_idx, end_idx) = get_square_stim_characteristics(current, t, no_test_pulse=True)
    stim_len = end_idx - start_idx

    distribution_start_ind=start_idx + int(.5/dt)
    distribution_end_ind=start_idx + stim_len
    
    v_section=voltage[distribution_start_ind:distribution_end_ind]
    if MAKE_PLOT:
        times=np.arange(0, len(voltage))*dt
        plt.figure(figsize=(15, 11))    
        plt.subplot2grid((7,2), (0,0), colspan=2)
        plt.plot(times[distribution_start_ind:distribution_end_ind], v_section)
        plt.title('voltage for histogram')

    print v_section
    v_section=v_section-np.mean(v_section)  
    var_of_section=np.var(v_section)
    sv_for_expsymm=np.std(v_section)/np.sqrt(2)
    subthreshold_long_square_voltage_distribution=stats.norm(loc=0, scale=np.sqrt(var_of_section))
    
    #--autocorrelation
    tau_4AC=res*cap
    AC=autocorr(v_section-np.mean(v_section)) 
    ACtime=np.arange(0,len(AC))*dt

    #--fit autocorrelation with decaying exponential    
    (popt, pcov)= curve_fit(exp_decay, ACtime, AC, p0=[AC[0],tau_4AC])
    tau_from_AC=popt[1]
    
    if MAKE_PLOT:        
        plt.subplot2grid((7,2), (1,0), rowspan=3)    
        plt.hist(v_section, bins=50, normed=True, label='data')
        data_grid=np.arange(min(v_section), max(v_section), abs(min(v_section)-max(v_section))/100.)
        plt.plot(data_grid, subthreshold_long_square_voltage_distribution.pdf(data_grid), 'r', label='gauss with\nmeasured var')
        plt.plot(data_grid, expsymm_pdf(data_grid, sv_for_expsymm), 'm', lw=3, label='expsymm function')
        plt.xlabel('voltage (mV)')
        plt.title('Mean subtracted voltage hist')
        plt.legend()
        
        #--cumulative density function
        (h, edges)=np.histogram(v_section, bins=50)        
        centers=find_bin_center(edges)

        CDFx=centers
        CDFy=np.cumsum(h)/float(len(v_section))

        plt.subplot2grid((7,2), (4,0), rowspan=3)
        plt.plot(CDFx, CDFy, label='data')
#        plt.plot(CDFx, sig(CDFx, popt[0], popt[1]), label='fit')
        plt.plot(data_grid, subthreshold_long_square_voltage_distribution.cdf(data_grid), 'r', label='gauss with\nmeasured var')
        plt.plot(data_grid, expsymm_cdf(data_grid, sv_for_expsymm), 'm', lw=3, label='expsymm func')
        plt.title('Normalized cumulative sum')
        plt.xlabel('v-mean(v)')
        plt.legend()
        
        plt.subplot2grid((7,2), (1,1), rowspan=3)
        plt.plot(ACtime, AC, label='data')
        plt.xlabel('shift (s)')
        plt.title('Auto correlation')
        plt.plot(ACtime, exp_decay(ACtime, AC[0], tau_4AC), label='RC')
        plt.plot(ACtime, exp_decay(ACtime, popt[0], tau_from_AC), label='fit')
        plt.legend()
    
        plt.tight_layout()
        if SHOW_PLOT:
            plt.show(block=BLOCK)  

        if PUBLICATION_PLOT:     
            times=np.arange(0, len(voltage))*dt
            plt.figure(figsize=(14, 7))    
            plt.subplot2grid((3,3), (0,0), colspan=3)
            plt.xlabel('time (s)', fontsize=14)
            plt.ylabel('(mV)', fontsize=14)
            plt.plot(times[distribution_start_ind:distribution_end_ind], v_section*1.e3)
            plt.title('Voltage for histogram', fontsize=16)
               
            plt.subplot2grid((3,3), (1,0), rowspan=2)    
            plt.hist(v_section*1.e3, bins=50, normed=True, label='data')
            data_grid=np.arange(min(v_section), max(v_section), abs(min(v_section)-max(v_section))/100.)
#                    plt.plot(data_grid, subthreshold_long_square_voltage_distribution.pdf(data_grid), 'r', label='gauss with\nmeasured var')
            plt.plot(data_grid*1.e3, 1.e-3*expsymm_pdf(data_grid, sv_for_expsymm), 'm', lw=3, label='expsymm ')
            plt.xlabel('voltage (mV)', fontsize=14)
            plt.title('Mean subtracted voltage hist', fontsize=16)
            plt.legend(loc=1)
            
            #--cumulative density function
            (h, edges)=np.histogram(v_section, bins=50)
            centers=find_bin_center(edges)

            CDFx=centers
            CDFy=np.cumsum(h)/float(len(v_section))

            plt.subplot2grid((3,3), (1,1), rowspan=2)
            plt.plot(CDFx*1e3, CDFy, label='data')
    #        plt.plot(CDFx, sig(CDFx, popt[0], popt[1]), label='fit')
#                    plt.plot(data_grid, subthreshold_long_square_voltage_distribution.cdf(data_grid), 'r', label='gauss with\nmeasured var')
            plt.plot(data_grid*1.e3, expsymm_cdf(data_grid, sv_for_expsymm), 'm', lw=3, label='expsymm')
            plt.title('Normalized cumulative sum',  fontsize=16)
            plt.xlabel('V-mean(V) (mV)', fontsize=16)
            plt.legend(loc=2, fontsize=14)
            
            plt.subplot2grid((3,3), (1,2), rowspan=2)
            plt.plot(ACtime, AC*1.e3, label='data')
            plt.xlabel('shift (s)', fontsize=14)
            plt.title('Auto correlation',  fontsize=16)
#                    plt.plot(ACtime, exp_decay(ACtime, AC[0], tau_4AC), label='RC')
            plt.plot(ACtime, exp_decay(ACtime, popt[0]*1.e3, tau_from_AC), lw=3, label='fit')
            plt.legend(loc=1)
            plt.tight_layout()
            
    return var_of_section, sv_for_expsymm, tau_from_AC

def expsymm_pdf(v, dv):
    return 1./(2.*dv)*np.exp(-np.absolute(v)/dv)

def expsymm_cdf(v, dv):
    return 1./2.+(v*(1-np.exp(-np.absolute(v)/dv)))/(2.*np.absolute(v))

def exp_decay(time, amp, tau):
    return amp*np.exp(-time/tau) 

def find_bin_center(edges):
    centers=np.zeros(len(edges)-1)
    for ii in range(0, len(edges)-1):
        centers[ii]=np.mean([edges[ii], edges[ii+1]])
    return centers

def autocorr(x):
    result = np.correlate(x, x, mode='full')
#    return result
    return result[result.size/2:]



from __future__ import division

import SimpleITK as sitk
import numpy as np


def convert_axis(axis):
    return 2 - axis


def max_projection(volume, axis, *a, **k):
    volume = sitk.GetArrayFromImage(volume)
    axis = convert_axis(axis)

    return np.amax(volume, axis), np.argmax(volume, axis)


def template_projection(volume, axis, gain=2, maxv=1, *a, **k):
    volume = sitk.GetArrayFromImage(volume)
    axis = convert_axis(axis)

    output_shape = list(volume.shape)
    del output_shape[axis]
    output = np.zeros(output_shape, dtype=float)

    for ii in xrange(volume.shape[axis]):
        current = volume.take(ii, axis)

        output = np.multiply(output, (maxv - current) / maxv)
        output += gain * np.multiply(current, current) / maxv

    return output






###
# Purpose:
#
#   Generates the heatmap files for GBM analysis runs. The files are the following:
#   transcripts_for_genes.csv, genes_for_transcripts.csv, gene_fpkm_table.csv, transcript_fpkm_table.csv
#
# Usage:
#
#   python generate_gbm_heatmap.py input.json
#
# Input:
#
#   input.json
#
#   {
#       "transcripts_for_genes_output":"/tmp/transcripts_for_genes.csv",
#       "genes_for_transcripts_output":"/tmp/genes_for_transcripts.csv",
#       "gene_fpkm_table_output":"/tmp/gene_fpkm_table.csv",
#       "transcript_fpkm_table_output":"/tmp/transcript_fpkm_table.csv",
#       "columns_samples_output":"/tmp/columns_samples.csv",
#       "analysis_run_records":"/tmp/analysis_run_records.json",
#       "sample_metadata_records":"/tmp/sample_metadata_records.json"
#   }
#
# Output:
#
#   Creates the specified csv files

import json
import sys
import numpy as np
import pandas as pd


def create_transcripts_for_genes(analysis_run_gene_file):

    """ Creates a list that contains the associated transcript for each gene sorted by entrez_id """

    transcripts_for_genes = np.genfromtxt(analysis_run_gene_file["analysis_run_gene_path"], usecols=[0, 1], skip_header=1,
                                          dtype='str').tolist()
    data = sorted(transcripts_for_genes, key=lambda row: int(row[0]))
    header = ['gene_id', 'transcript_id(s)']
    data.insert(0, header)
    data = pd.DataFrame(data)
    return data


def create_genes_for_transcripts(analysis_run_transcript_file):

    """ Creates a list that contains the associated gene for each transcript sorted alphabetically """

    genes_for_transcripts = np.genfromtxt(analysis_run_transcript_file["analysis_run_transcript_path"], usecols=[0, 1]
                                          , skip_header=1, dtype='str').tolist()
    data = sorted(genes_for_transcripts, key=lambda row: row[0].lower())
    header = ['transcript_id', 'gene_id']
    data.insert(0, header)
    data = pd.DataFrame(data)
    return data


def create_gene_fpkm_table(analysis_run_records):

    """ Creates a a matrix ("rows x columns = genes x samples") of fpkm gene expression values for each particular
        (gene, sample) pair. Rows are sorted by entrez_id and columns are by rna_well_id """

    gene_fpkm = []
    rna_well_ids = []

    for record in analysis_run_records:
        gene_fpkm.append(np.genfromtxt(record["analysis_run_gene_path"], usecols=[-1]
                                       , skip_header=1, dtype='str'))
        rna_well_ids.append(record["rna_well_id"])

    entrez_ids = np.genfromtxt(analysis_run_records[0]["analysis_run_gene_path"], usecols=[0], skip_header=1
                               , dtype='str').tolist()
    entrez_ids_int = list(map(int, entrez_ids))
    gene_fpkm_table = np.column_stack(gene_fpkm)
    df = pd.DataFrame(gene_fpkm_table, columns=rna_well_ids, index=entrez_ids_int)
    df = df.sort_index()
    rna_well_ids_sorted = sorted(list(map(int, rna_well_ids)))
    data = df[rna_well_ids_sorted]
    return data


def create_transcript_fpkm_table(analysis_run_records):

    """ Creates a a matrix ("rows x columns = transcripts x samples") of fpkm gene expression values for each particular
        (transcript, sample) pair. Rows are sorted by transcript id and columns are by rna_well_id """

    transcript_fpkm = []
    rna_well_ids = []

    for record in analysis_run_records:
        transcript_fpkm.append(np.genfromtxt(record["analysis_run_transcript_path"], usecols=[-1]
                                             , skip_header=1, dtype='str'))
        rna_well_ids.append(record["rna_well_id"])

    transcript_ids = np.genfromtxt(analysis_run_records[0]["analysis_run_transcript_path"], usecols=[0], skip_header=1
                                   , dtype='str').tolist()
    transcript_fpkm_table = np.column_stack(transcript_fpkm)

    df = pd.DataFrame(transcript_fpkm_table, columns=rna_well_ids, index=transcript_ids)
    df = df.sort_index()
    rna_well_ids_sorted = sorted(list(map(int, rna_well_ids)))
    data = df[rna_well_ids_sorted]
    return data


def create_sample_metadata(sample_metadata_records):

    """ Creates a table of sample metadata sorted by rna_well_id """

    df = pd.DataFrame.from_dict(sample_metadata_records, orient='columns')
    data = df.sort_values(by=['rna_well_id']).reset_index(drop=True)
    rna_well_id = data['rna_well_id']
    data.drop(labels=['rna_well_id'], axis=1, inplace=True)
    data.insert(0, 'rna_well_id', rna_well_id)
    return data


def main():

    input_file = sys.argv[1]
    data = json.load(open(input_file))
    transcripts_for_genes_output = data["transcripts_for_genes_output"]
    genes_for_transcripts_output = data["genes_for_transcripts_output"]
    gene_fpkm_table_output = data["gene_fpkm_table_output"]
    transcript_fpkm_table_output = data["transcript_fpkm_table_output"]
    columns_samples_output = data["columns_samples_output"]
    analysis_run_records = json.load(open(data["analysis_run_records"]))
    sample_metadata_records = json.load(open(data["sample_metadata_records"]))

    create_transcripts_for_genes(analysis_run_records["analysis_run_records"][0]).to_csv(transcripts_for_genes_output,
                                                                                         index=False, header=False)
    create_genes_for_transcripts(analysis_run_records["analysis_run_records"][0]).to_csv(genes_for_transcripts_output,
                                                                                         index=False, header=False)
    create_gene_fpkm_table(analysis_run_records["analysis_run_records"]).to_csv(gene_fpkm_table_output)
    create_transcript_fpkm_table(analysis_run_records["analysis_run_records"]).to_csv(transcript_fpkm_table_output)
    create_sample_metadata(sample_metadata_records).to_csv(columns_samples_output, index=False)


if __name__ == '__main__':
    main()



import logging
import os
import functools

import SimpleITK as sitk
from scipy.misc import imread, imsave
import numpy as np
import pandas as pd

from allensdk.internal.core.lims_pipeline_module import PipelineModule

from allensdk.internal.mouse_connectivity.projection_thumbnail.volume_utilities import sitk_get_diagonal_length
from allensdk.internal.mouse_connectivity.projection_thumbnail.generate_projection_strip import run, apply_colormap
from allensdk.internal.mouse_connectivity.projection_thumbnail.visualization_utilities import convert_discrete_colormap


PERMUTATION = [2, 1, 0]
FLIP = [True, False, False]


def write_depth_image(image, path):
    image = sitk.GetImageFromArray(image)
    sitk.WriteImage(image, str(path))
    

def load_background_image(path):
    background = sitk.ReadImage(str(path))
    background = sitk.GetArrayFromImage(background)
    bg_split = np.split(background, 3, axis=-1)
    return bg_split[0] / 255.0


def no_pad(volume):
    shape = [int(np.ceil(sitk_get_diagonal_length(volume))), 0, 0]
    shape[1] = volume.GetSize()[1]
    shape[2] = volume.GetSize()[2]
    return shape


def pad(volume):
    shape = [int(np.ceil(sitk_get_diagonal_length(volume))), 0, 0]
    shape[1] = int(np.floor(np.linalg.norm([volume.GetSize()[1], volume.GetSize()[0]])))
    shape[2] = int(np.floor(np.linalg.norm([volume.GetSize()[2], volume.GetSize()[0]])))
    return shape


def main():

    module = PipelineModule()
    input_data = module.input_data()

    output_dir = os.path.dirname(module.args.output_json)

    logging.info('reading data volume from {0}'.format(input_data['volume_path']))
    volume = sitk.ReadImage(str(input_data['volume_path']))
    volume = sitk.PermuteAxes(volume, PERMUTATION)
    volume = sitk.Flip(volume, FLIP)

    logging.info('reading colormap from {0}'.format(input_data['colormap_path']))
    colormap = pd.read_csv(input_data['colormap_path'], header=None, 
                           names=['red', 'green', 'blue'], delim_whitespace=True)
    colormap = convert_discrete_colormap(colormap.values, 'projection')

    output_data = {'output_file_paths': []}
    for rot in input_data['rotations']:
    
        rot['write_depth_sheet'] = functools.partial(write_depth_image, 
                                                     path=str(os.path.join(output_dir, rot['depth_path'])))
        output_data['output_file_paths'].append(os.path.join(output_dir, rot['depth_path']))

        if isinstance(rot['window_size'], (str, unicode)):
            if rot['window_size'] == 'no_pad':
                rot['window_size'] = no_pad(volume)
            elif rot['window_size'] == 'pad':
                rot['window_size'] = pad(volume)
            else:
                raise ValueError('did not understand window size option {0}'.format(rot['window_size']))
        logging.info('window_size: {0}'.format(rot['window_size']))
        
        for out_image in rot['output_images']:
            out_image['write'] = functools.partial(imsave, os.path.join(output_dir, out_image['path']))
            output_data['output_file_paths'].append(os.path.join(output_dir, out_image['path']))            

            if 'background_path' in out_image:
                out_image['background'] = load_background_image(out_image['background_path'])
            else:
                out_image['background'] = None

    run(volume, input_data['min_threshold'], input_data['max_threshold'], 
        input_data['rotations'], colormap)
    module.write_output_data(output_data)


if __name__ == '__main__':
    main()



import numpy as np
import h5py


def open_view_on_binary(file_like, dtype=np.uint8, mode="r", offset=0,
                        shape=None, order="C", strides=None):
    '''Open a view into a memory-mapped binary file.

    Parameters
    ----------
    file_like : {string, file object}
        File to open.
    dtype : numpy.dtype
        Numpy dtype to open the memory-mapped array as.
    mode : string
        Mode to open the file in.
    offset : integer
        Offset (in bytes) into the file at which to start the memory
        map.
    shape : {tuple, list}
        Shape of the array.
    order : {"C", "F"}
        C or Fortran ordering.
    strides : {tuple, list}
        Strides along each axis for reading the array.

    Returns
    -------
    numpy.memmap
        Strided view into memory-mapped array.
    '''
    mapped = np.memmap(file_like, dtype, mode, offset, order=order)
    return np.lib.stride_tricks.as_strided(mapped, shape=shape,
                                           strides=strides)


def read_strided(filename, dtype, offset, shape, strides):
    '''Load a frame without memory-mapping.'''
    frame_size = np.dtype(dtype).itemsize
    arr = np.empty(shape, dtype=dtype)
    frame_size = arr.dtype.itemsize*np.product(shape[1:])
    step = strides[0] - frame_size
    with open(filename, "rb") as f:
        f.seek(offset)
        for i in range(shape[0]):
            frame = np.frombuffer(f.read(frame_size), dtype=dtype)
            arr[i] = frame.reshape(shape[1:])
            f.seek(step, 1)
    return arr


def load_frame(raw_filename, json_meta, use_memmap=False):
    '''Load a frame of a multi-frame raw file.'''
    if use_memmap:
        arr = open_view_on_binary(raw_filename, dtype=json_meta["dtype"],
                                  offset=json_meta["byte_offset"],
                                  shape=json_meta["shape"],
                                  strides=json_meta["strides"])
    else:
        arr = read_strided(raw_filename, dtype=json_meta["dtype"],
                           offset=json_meta["byte_offset"],
                           shape=json_meta["shape"],
                           strides=json_meta["strides"])
    return arr


def export_frame_to_hdf5(raw_filename, data_hdf5_filename,
                         auxiliary_hdf5_filename, frame_meta,
                         compression="gzip", compression_opts=9):
    '''Export a frame from raw to hdf5.
    
    Data with the channel_description `data` is stored in the
    data_hdf5_filename, while any other data is stored in the
    auxiliary_hdf5_filename 
    '''
    data_created = False
    aux_created = False
    for json_meta in frame_meta:
        # This is dirty, we should expand the metadata handoff to allow
        # real specification of ophys data versus auxiliary data
        mode = "a"
        if json_meta["channel_description"] == "data":
            filename = data_hdf5_filename
            if not data_created:
                data_created = True
                mode = "w"
        else:
            filename = auxiliary_hdf5_filename
            if not aux_created:
                aux_created = True
                mode = "w"
        data = load_frame(raw_filename, json_meta)
        chunks = (1, json_meta["shape"][1], json_meta["shape"][2])
        with h5py.File(filename, mode) as f:
            f.create_dataset(json_meta["channel_description"], data=data,
                             chunks=chunks, compression=compression,
                             compression_opts=compression_opts)



class SimpleTree( object ):
    def __init__(self, nodes, 
                 node_id_cb, 
                 parent_id_cb):

        self.node_list = nodes

        self._nodes = { node_id_cb(n):n for n in nodes }
        self._parent_ids = { nid:parent_id_cb(n) for nid,n in self._nodes.iteritems() }
        self._child_ids = { nid:[] for nid in self._nodes }

        for nid in self._parent_ids:
            pid = self._parent_ids[nid]
            if pid:
                self._child_ids[pid].append(nid)

    def node_ids(self):
        return self._nodes.keys()

    def parent_id(self, nid):
        try:
            return self._parent_ids[nid]
        except KeyError:
            raise KeyError("Could not find parent for node %s" % str(nid))

    def child_ids(self, nid):
        try:
            for cid in self._child_ids[nid]:
                yield cid
        except KeyError:
            raise KeyError("Could not find children for node %s" % str(nid))

    def ancestor_ids(self, nid):
        try:
            pid = nid
            while pid:
                yield pid
                pid = self.parent_id(pid)
        except:
            raise KeyError("Could not find ancestors for node %s" % str(nid))
        
    def descendant_ids(self, nid):
        ids = [nid]
        try:
            while ids:
                nid = ids.pop()
                yield nid
                ids += self.child_ids(nid)
        except KeyError:
            raise KeyError("Could not find descendants for node %s" % str(nid))

    def node(self, nid):
        return self._nodes[nid]

    def nodes(self, nids=None):
        if nids is None:
            nids = self.node_ids()

        for nid in nids:
            yield self.node(nid)

    def parent(self, nid):
        return self.node(self.parent_id(nid))

    def children(self, nid):
        for node in self.nodes(self.child_ids(nid)):
            yield node

    def descendants(self, nid):
        for node in self.nodes(self.descendant_ids(nid)):
            yield node

    def ancestors(self, nid):
        for node in self.nodes(self.ancestor_ids(nid)):
            yield node


        

